import scri
import pytest
import numpy as np
import spinsfast
import spherical_functions as sf
import quaternion
from quaternion.calculus import derivative

ABD = scri.AsymptoticBondiData


def kerr_schild(mass, spin, ell_max=8):
    psi2 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)
    psi1 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)
    psi0 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)

    # In the Moreschi-Boyle convention
    psi2[0] = -sf.constant_as_ell_0_mode(mass)
    psi1[2] = -np.sqrt(2) * (3j * spin / 2) * (np.sqrt((8 / 3) * np.pi))
    psi0[6] = 2 * (3 * spin**2 / mass / 2) * (np.sqrt((32 / 15) * np.pi))

    return psi2, psi1, psi0


def test_abd_to_abd():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=5000)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    supertranslation = np.array(
        [0.0, 3e-2 - 1j * 5e-3, 1e-3, -3e-2 - 1j * 5e-3, 2e-4 + 1j * 1e-4, 1j * 3e-3, 1e-2, 1j * 3e-3, 2e-4 - 1j * 1e-4]
    )
    frame_rotation = quaternion.quaternion(1, 2, 3, 4).normalized().components
    boost_velocity = np.array([2e-4, -3e-4, -5e-4])

    abd_target = abd.transform(
        supertranslation=supertranslation, frame_rotation=frame_rotation, boost_velocity=boost_velocity
    )

    # We don't perform time/phase fixing because this data is radiation-free so it doesn't make sense
    abd_recovered, transformation, rel_err = abd.map_to_abd_frame(
        abd_target,
        t_0=0,
        padding_time=20,
        N_itr_maxes={"abd": 2, "superrest": 1, "CoM_transformation": 10, "rotation": 10, "supertranslation": 10},
        fix_time_phase_freedom=False,
        nprocs=-1,
    )

    abd_target_interp = abd_target.interpolate(
        abd_target.t[
            np.argmin(abs(abd_target.t - max(abd_target.t[0], abd_recovered.t[0]))) : np.argmin(
                abs(abd_target.t - min(abd_target.t[-1], abd_recovered.t[-1]))
            )
            + 1
        ]
    )
    abd_recovered_interp = abd_recovered.interpolate(
        abd_recovered.t[
            np.argmin(abs(abd_recovered.t - max(abd_target.t[0], abd_recovered.t[0]))) : np.argmin(
                abs(abd_recovered.t - min(abd_target.t[-1], abd_recovered.t[-1]))
            )
            + 1
        ]
    )

    assert np.allclose(
        np.array(
            [
                abd_target_interp.sigma,
                abd_target_interp.psi4,
                abd_target_interp.psi3,
                abd_target_interp.psi2,
                abd_target_interp.psi1,
                abd_target_interp.psi0,
            ]
        ),
        np.array(
            [
                abd_recovered_interp.sigma,
                abd_recovered_interp.psi4,
                abd_recovered_interp.psi3,
                abd_recovered_interp.psi2,
                abd_recovered_interp.psi1,
                abd_recovered_interp.psi0,
            ]
        ),
    )



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the AsymptoticBondData
### class.  In particular, they assume that the first argument, `self` is an instance of
### AsymptoticBondData.  They should probably not be used outside of that class.

import scri
import numpy as np
from math import sqrt, pi
import spherical_functions as sf


def mass_aspect(self, truncate_ell=max):
    """Compute the Bondi mass aspect of the AsymptoticBondiData.

    The Bondi mass aspect is given by

        M = -ℜ{ψ₂ + σ ∂ₜσ̄}

    Note that the last term is a product between two fields.  If, for example, these both have
    ell_max=8, then their full product would have ell_max=16, meaning that we would go from
    tracking 81 modes to 289.  This shows that deciding how to truncate the output ell is
    important, which is why this function has the extra argument that it does.

    Parameters
    ==========
    truncate_ell: int, or callable [defaults to `max`]
        Determines how the ell_max value of the output is determined.  If an integer is passed,
        each term in the output is truncated to have at most that ell_max.  (In particular,
        terms that will not be used in the output are simply not computed, without incurring any
        errors due to aliasing.)  If a callable is passed, it is passed on to the
        spherical_functions.Modes.multiply method.  See that function's docstring for details.
        The default behavior will result in the output having ell_max equal to the largest of
        any of the individual Modes objects in the equation for M above -- but not the
        product.

    """
    if callable(truncate_ell):
        return -(self.psi2 + self.sigma.multiply(self.sigma.bar.dot, truncator=truncate_ell)).real
    elif truncate_ell:
        return -(
            self.psi2.truncate_ell(truncate_ell)
            + self.sigma.multiply(self.sigma.bar.dot, truncator=lambda tup: truncate_ell)
        ).real
    else:
        return -(self.psi2 + self.sigma * self.sigma.bar.dot).real


def charge_vector_from_aspect(charge):
    """Output the ell<=1 modes of a BMS charge aspect as the charge four-vector.

    Considering the aspect as a function a(θ, ϕ), we express the corresponding
    four-vector in terms of Cartesian components as

      vᵝ = (1/4π) ∫ ℜ{a} (tᵝ + rᵝ) dΩ

    where the integral is taken over the 2-sphere, and tᵝ + rᵝ has components
    (1, sinθ cosϕ, sinθ sinϕ, cosθ).

    """
    four_vector = np.empty(charge.shape, dtype=float)
    four_vector[..., 0] = charge[..., 0].real
    four_vector[..., 1] = (charge[..., 1] - charge[..., 3]).real / sqrt(6)
    four_vector[..., 2] = (charge[..., 1] + charge[..., 3]).imag / sqrt(6)
    four_vector[..., 3] = charge[..., 2].real / sqrt(3)
    return four_vector / np.sqrt(4 * np.pi)


def bondi_rest_mass(self):
    """Compute the rest mass from the Bondi four-momentum"""
    four_momentum = self.bondi_four_momentum()
    rest_mass = np.sqrt(four_momentum[:, 0] ** 2 - np.sum(four_momentum[:, 1:] ** 2, axis=1))
    return rest_mass


def bondi_four_momentum(self):
    """Compute the Bondi four-momentum

    This is just the ell<2 component of the mass aspect, expressed as a four-vector.

    """
    ell_max = 1  # Compute only the parts we need, ell<=1
    charge_aspect = self.mass_aspect(ell_max).view(np.ndarray)
    return charge_vector_from_aspect(charge_aspect)


def bondi_angular_momentum(self):
    """Compute the total Bondi angular momentum vector

        i (ψ₁ + σ ðσ̄)

    See Eq. (8) of Dray (1985) iopscience.iop.org/article/10.1088/0264-9381/2/1/002

    """
    ell_max = 1  # Compute only the parts we need, ell<=1
    charge_aspect = (
        1j
        * (self.psi1.truncate_ell(ell_max) + self.sigma.multiply(self.sigma.bar.eth_GHP, truncator=lambda tup: ell_max))
    ).view(np.ndarray)
    return charge_vector_from_aspect(charge_aspect)[:, 1:]


def CWWY_angular_momentum(self):
    """Compute the Chen/Wang/Wang/Yau angular momentum vector.

    See Eq. (5) of https://arxiv.org/abs/2102.03235

    """
    ell_max = 1  # Compute only the parts we need, ell<=1
    supertranslation_potential_data = scri.asymptotic_bondi_data.map_to_superrest_frame.𝔇inverse(
        np.array(self.sigma.ethbar_GHP.ethbar_GHP + self.sigma.bar.eth_GHP.eth_GHP), self.ell_max
    )
    supertranslation_potential = scri.ModesTimeSeries(
        sf.SWSH_modes.Modes(
            supertranslation_potential_data,
            spin_weight=0,
            ell_min=0,
            ell_max=self.ell_max,
            multiplication_truncator=max,
        ),
        time=self.t,
    )
    charge_aspect = (
        1j
        * (
            self.psi1.truncate_ell(ell_max)
            + self.sigma.multiply(self.sigma.bar.eth_GHP, truncator=lambda tup: ell_max)
            + supertranslation_potential.multiply(self.mass_aspect().eth_GHP, truncator=lambda tup: ell_max)
        )
    ).view(np.ndarray)
    return charge_vector_from_aspect(charge_aspect)[:, 1:]


def bondi_dimensionless_spin(self):
    """Compute the dimensionless Bondi spin vector"""
    N = self.bondi_boost_charge()
    J = self.bondi_angular_momentum()
    P = self.bondi_four_momentum()
    M_sqr = (P[:, 0] ** 2 - np.sum(P[:, 1:] ** 2, axis=1))[:, np.newaxis]
    v = P[:, 1:] / (P[:, 0])[:, np.newaxis]
    v_norm = np.linalg.norm(v, axis=1)
    # To prevent dividing by zero, we compute the normalized velocity vhat ONLY at
    # timesteps with a non-zero velocity.
    vhat = v.copy()
    t_idx = v_norm != 0  # Get the indices for timesteps with non-zero velocity
    vhat[t_idx] = v[t_idx] / v_norm[t_idx, np.newaxis]
    gamma = (1 / np.sqrt(1 - v_norm**2))[:, np.newaxis]
    J_dot_vhat = np.einsum("ij,ij->i", J, vhat)[:, np.newaxis]
    spin_charge = (gamma * (J + np.cross(v, N)) - (gamma - 1) * J_dot_vhat * vhat) / M_sqr
    return spin_charge


def bondi_boost_charge(self):
    """Compute the Bondi boost charge vector

        - [ψ₁ + σ ðσ̄ + ½ð(σ σ̄) - t ð ℜ{ψ₂ + σ ∂ₜσ̄}]

    See Eq. (8) of Dray (1985) iopscience.iop.org/article/10.1088/0264-9381/2/1/002

    """
    ell_max = 1  # Compute only the parts we need, ell<=1
    charge_aspect = -(
        self.psi1.truncate_ell(ell_max)
        + self.sigma.multiply(self.sigma.bar.eth_GHP, truncator=lambda tup: ell_max)
        + 0.5 * (self.sigma.multiply(self.sigma.bar, truncator=lambda tup: ell_max)).eth_GHP
        - self.t
        * (
            self.psi2.truncate_ell(ell_max) + self.sigma.multiply(self.sigma.bar.dot, truncator=lambda tup: ell_max)
        ).real.eth_GHP
    ).view(np.ndarray)
    return charge_vector_from_aspect(charge_aspect)[:, 1:]


def bondi_CoM_charge(self):
    """Compute the center-of-mass charge vector

        Gⁱ = Nⁱ + t Pⁱ = - [ψ₁ + σ ðσ̄ + ½ð(σ σ̄)]

    where Nⁱ is the boost charge and Pⁱ is the momentum.  See Eq. (3.4) of arXiv:1912.03164.

    """
    ell_max = 1  # Compute only the parts we need, ell<=1
    charge_aspect = -(
        self.psi1.truncate_ell(ell_max)
        + self.sigma.multiply(self.sigma.bar.eth_GHP, truncator=lambda tup: ell_max)
        + 0.5 * (self.sigma.multiply(self.sigma.bar, truncator=lambda tup: ell_max)).eth_GHP
    ).view(np.ndarray)
    return charge_vector_from_aspect(charge_aspect)[:, 1:]


def supermomentum(self, supermomentum_def, **kwargs):
    """Compute the supermomentum

    This function allows for several different definitions of the
    supermomentum.  These differences only apply to ell > 1 modes,
    so they do not affect the Bondi four-momentum.  See
    Eqs. (7-9) in arXiv:1404.2475 for the different supermomentum
    definitions and links to further references.

    In the literature, there is an ambiguity of vocabulary.  When
    it comes to other BMS charges, we clearly distinuish between
    the "charge" and the "aspect".  However, the
    term "supermomentum" is used for both.  Accordingly, this
    function provides two ways to compute the supermomentum.

    1) By default, the supermomentum will be computed as

        Ψ = ψ₂ + σ ∂ₜσ̄ + f(θ, ϕ)

       (See below for the definitions of `f`.)

    2) By passing the option `integrated=True`, the supermomentum
       will instead be computed as

        Pₗₘ = - (1/4π) ∫ Ψ(θ, ϕ) Yₗₘ(θ, ϕ) dΩ

    Parameters
    ----------
    supermomentum_def : str
        The definition of the supermomentum to be computed.  One of the
        following (case-insensitive) options can be specified:
          * 'Bondi-Sachs' or 'BS' for f = 0
          * 'Moreschi' or 'M' for f = ð²σ̄
          * 'Geroch' or 'G' for f = ½ (ð²σ̄ - ð̄²σ)
          * 'Geroch-Winicour' or 'GW' for f = - ð̄²σ
    integrated : bool, optional
        If True, then return the integrated form of the supermomentum — see
        Eq. (6) in arXiv:1404.2475.  Default is False
    working_ell_max: int, optional
        The value of ell_max to be used to define the computation grid. The
        number of theta points and the number of phi points are set to
        2*working_ell_max+1. Defaults to 2*self.ell_max.

    Returns
    -------
    ModesTimeSeries

    """
    return_integrated = kwargs.pop("integrated", False)

    if supermomentum_def.lower() in ["bondi-sachs", "bs"]:
        supermomentum = self.psi2 + self.sigma.grid_multiply(self.sigma.bar.dot, **kwargs)
    elif supermomentum_def.lower() in ["moreschi", "m"]:
        supermomentum = (
            self.psi2 + self.sigma.grid_multiply(self.sigma.bar.dot, **kwargs) + self.sigma.bar.eth_GHP.eth_GHP
        )
    elif supermomentum_def.lower() in ["geroch", "g"]:
        supermomentum = (
            self.psi2
            + self.sigma.grid_multiply(self.sigma.bar.dot, **kwargs)
            + 0.5 * (self.sigma.bar.eth_GHP.eth_GHP - self.sigma.ethbar_GHP.ethbar_GHP)
        )
    elif supermomentum_def.lower() in ["geroch-winicour", "gw"]:
        supermomentum = (
            self.psi2 + self.sigma.grid_multiply(self.sigma.bar.dot, **kwargs) - self.sigma.ethbar_GHP.ethbar_GHP
        )
    else:
        raise ValueError(
            f"Supermomentum defintion '{supermomentum_def}' not recognized. Please choose one of "
            "the following options:\n"
            "  * 'Bondi-Sachs' or 'BS'\n"
            "  * 'Moreschi' or 'M'\n"
            "  * 'Geroch' or 'G'\n"
            "  * 'Geroch-Winicour' or 'GW'"
        )
    if return_integrated:
        return -0.5 * supermomentum.bar / np.sqrt(np.pi)
    else:
        return supermomentum



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import sphinx

from recommonmark.parser import CommonMarkParser

# -- Project information -----------------------------------------------------

from scri import __version__ as scri_version

# -- Project information -----------------------------------------------------

project = "scri"
copyright = "2016, Michael Boyle"
author = "Michael Boyle"

# The short X.Y version
version = scri_version
# The full version, including alpha/beta/rc tags
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "recommonmark",
]

autosummary_generate = True

autodoc_docstring_signature = True
if sphinx.version_info < (1, 8):
    autodoc_default_flags = ["members", "undoc-members"]
else:
    autodoc_default_options = {"members": None, "undoc-members": None, "special-members": "__call__"}

# -- Try to auto-generate numba-decorated signatures -----------------

import numba
import inspect


def process_numba_docstring(app, what, name, obj, options, signature, return_annotation):
    if type(obj) is not numba.core.registry.CPUDispatcher:
        return (signature, return_annotation)
    else:
        original = obj.py_func
        orig_sig = inspect.signature(original)

        if (orig_sig.return_annotation) is inspect._empty:
            ret_ann = None
        else:
            ret_ann = orig_sig.return_annotation.__name__

        return (str(orig_sig), ret_ann)


def setup(app):
    app.connect("autodoc-process-signature", process_numba_docstring)


# --------------------------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
if sphinx.version_info < (1, 8):
    source_parsers = {
        ".md": CommonMarkParser,
    }
    source_suffix = [".rst", ".md"]
else:
    source_suffix = {
        ".rst": "restructuredtext",
        ".txt": "markdown",
        ".md": "markdown",
    }

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import math
import numpy as np
import quaternion
import spherical_functions as sf
from .waveform_base import waveform_alterations
from .mode_calculations import corotating_frame, angular_velocity, LLDominantEigenvector
from .utilities import transition_function
from . import jit, Coprecessing, Coorbital, Corotating, Inertial


@waveform_alterations
def to_coprecessing_frame(W, RoughDirection=np.array([0.0, 0.0, 1.0]), RoughDirectionIndex=None, transition_times=None):
    """Transform waveform (in place) to a coprecessing frame

    Parameters
    ==========
    W: waveform
        Waveform object to be transformed in place.
    RoughDirection: 3-array [defaults to np.array([0.0, 0.0, 1.0])]
        Vague guess about the preferred initial axis, to choose the sign of the
        eigenvectors.
    RoughDirectionIndex: int or None [defaults to None]
        Time index at which to apply RoughDirection guess.
    transition_times : 2-tuple or None [defaults to None]
        If a 2-tuple of floats is given, these will be interpreted as the
        beginning and ending times (respectively) to transition from using the
        coprecessing frame to stopping rotation altogether.  This can be helpful
        for ensuring that the frame doesn't fluctuate wildly during ringdown.

    """
    if RoughDirectionIndex is None:
        RoughDirectionIndex = W.n_times // 8
    dpa = LLDominantEigenvector(W, RoughDirection=RoughDirection, RoughDirectionIndex=RoughDirectionIndex)
    R = np.array([quaternion.quaternion.sqrt(-quaternion.quaternion(0, *q).normalized() * quaternion.z) for q in dpa])
    R = quaternion.minimal_rotation(R, W.t, iterations=3)
    if transition_times is not None:
        i0, i1 = np.argmin(np.abs(W.t-transition_times[0])), np.argmin(np.abs(W.t-transition_times[1]))
        transition = transition_function(W.t[i0:], W.t[i0], W.t[i1], y0=1.0, y1=0.0)
        omega = quaternion.angular_velocity(R[i0:], W.t[i0:]) * transition[:, np.newaxis]
        _, slowingR = quaternion.integrate_angular_velocity((W.t[i0:], omega), t0=W.t[i0], t1=W.t[-1], R0=R[i0])
        R = np.concatenate((R[:i0], slowingR))
    W.rotate_decomposition_basis(R)
    W._append_history(f"{W}.to_coprecessing_frame({RoughDirection}, {RoughDirectionIndex}, {transition_times})")
    W.frameType = Coprecessing
    return W


@waveform_alterations
def to_corotating_frame(
    W, R0=quaternion.one, tolerance=1e-12, z_alignment_region=None, return_omega=False, truncate_log_frame=False
):
    """Transform waveform (in place) to a corotating frame

    Parameters
    ==========
    W: waveform
        Waveform object to be transformed in place
    R0: quaternion [defaults to 1]
        Initial value of frame when integrating angular velocity
    tolerance: float [defaults to 1e-12]
        Absolute tolerance used in integration of angular velocity
    z_alignment_region: None or 2-tuple of floats [defaults to None]
        If not None, the dominant eigenvector of the <LL> matrix is aligned with the z axis,
        averaging over this portion of the data.  The first and second elements of the input are
        considered fractions of the inspiral at which to begin and end the average.  For example,
        (0.1, 0.9) would lead to starting 10% of the time from the first time step to the max norm
        time, and ending at 90% of that time.
    return_omega: bool [defaults to False]
        If True, return a 2-tuple consisting of the waveform in the corotating frame (the usual
        returned object) and the angular-velocity data.  That is frequently also needed, so this is
        just a more efficient way of getting the data.
    truncate_log_frame: bool [defaults to False]
        If True, set bits of log(frame) with lower significance than `tolerance` to zero, and use
        exp(truncated(log(frame))) to rotate the waveform.  Also returns `log_frame` along with the
        waveform (and optionally `omega`)

    """
    frame, omega = corotating_frame(
        W, R0=R0, tolerance=tolerance, z_alignment_region=z_alignment_region, return_omega=True
    )
    if truncate_log_frame:
        log_frame = quaternion.as_float_array(np.log(frame))
        power_of_2 = 2 ** int(-np.floor(np.log2(2 * tolerance)))
        log_frame = np.round(log_frame * power_of_2) / power_of_2
        frame = np.exp(quaternion.as_quat_array(log_frame))
    W.rotate_decomposition_basis(frame)
    W._append_history(
        f"{W}.to_corotating_frame({R0}, {tolerance}, {z_alignment_region}, {return_omega}, {truncate_log_frame})"
    )
    W.frameType = Corotating
    if return_omega:
        if truncate_log_frame:
            return (W, omega, log_frame)
        else:
            return (W, omega)
    else:
        if truncate_log_frame:
            return (W, log_frame)
        else:
            return W


@waveform_alterations
def to_inertial_frame(W):
    W.rotate_decomposition_basis(~W.frame)
    W._append_history(f"{W}.to_inertial_frame()")
    W.frameType = Inertial
    return W


def get_alignment_of_decomposition_frame_to_modes(w, t_fid, nHat_t_fid=quaternion.x, ell_max=None):
    """Find the appropriate rotation to fix the attitude of the corotating frame.

    This function simply finds the rotation necessary to align the
    corotating frame to the waveform at the fiducial time, rather than
    applying it.  This is called by `AlignDecompositionFrameToModes`
    and probably does not need to be called directly; see that
    function's documentation for more details.

    Parameters
    ----------
    w: WaveformModes
        Object to be aligned
    t_fid: float
        Fiducial time at which the alignment should happen
    nHat_t_fid: quaternion (optional)
        The approximate direction of nHat at t_fid; defaults to x
    ell_max: int
       Highest ell mode to use in computing the <LL> matrix

    """
    # We seek that R_c such that R_corot(t_fid)*R_c rotates the z axis
    # onto V_f.  V_f measured in this frame is given by
    #     V_f = R_V_f * Z * R_V_f.conjugate(),
    # (note Z rather than z) where R_V_f is found below.  But
    #     Z = R_corot * z * R_corot.conjugate(),
    # so in the (x,y,z) frame,
    #     V_f = R_V_f * R_corot * z * R_corot.conjugate() * R_V_f.conjugate().
    # Now, this is the standard composition for rotating physical
    # vectors.  However, rotation of the basis behaves oppositely, so
    # we want R_V_f as our constant rotation, applied as a rotation of
    # the decomposition basis.  We also want to rotate so that the
    # phase of the (2,2) mode is zero at t_fid.  This can be achieved
    # with an initial rotation.

    if ell_max is None:
        ell_max = w.ell_max

    if w.frameType not in [Coprecessing, Coorbital, Corotating]:
        message = (
            "get_alignment_of_decomposition_frame_to_modes only takes Waveforms in the "
            + "coprecessing, coorbital, or corotating frames.  This Waveform is in the "
            + "'{0}' frame."
        )
        raise ValueError(message.format(w.frame_type_string))

    if w.frame.size != w.n_times:
        message = (
            "get_alignment_of_decomposition_frame_to_modes requires full information about the Waveform's frame."
            + "This Waveform has {0} time steps, but only {1} rotors in its frame."
        )
        raise ValueError(message.format(w.n_times, np.asarray(w.frame).size))

    if t_fid < w.t[0] or t_fid > w.t[-1]:
        message = "The requested alignment time t_fid={0} is outside the range of times in this waveform ({1}, {2})."
        raise ValueError(message.format(t_fid, w.t[0], w.t[-1]))

    # Get direction of angular-velocity vector near t_fid
    i_t_fid = (w.t <= t_fid).nonzero()[0][-1]  # Find the largest index i with t[i-1] <= t_fid
    if i_t_fid < w.t.size - 1:
        i_t_fid += 1
    i1 = 0 if i_t_fid - 5 < 0 else i_t_fid - 5
    i2 = w.t.size if i1 + 11 > w.t.size else i1 + 11
    Region = w[i1:i2, 2].copy().to_inertial_frame()
    omegaHat = quaternion.quaternion(0, *(angular_velocity(Region)[i_t_fid - i1])).normalized()

    # omegaHat contains the components of that vector relative to the
    # inertial frame.  To get its components in this Waveform's
    # (possibly rotating) frame, we need to rotate it by the inverse
    # of this Waveform's `frame` data:
    if w.frame.size > 1:
        R = w.frame[i_t_fid]
        omegaHat = R.inverse() * omegaHat * R
    elif w.frame.size == 1:
        R = w.frame[0]
        omegaHat = R.inverse() * omegaHat * R

    # Interpolate the Waveform to t_fid
    Instant = w[i1:i2].copy().interpolate(np.array([t_fid,]))
    R_f0 = Instant.frame[0]

    # V_f is the dominant eigenvector of <LL>, suggested by O'Shaughnessy et al.
    V_f = quaternion.quaternion(0, *(LLDominantEigenvector(Instant[:, : ell_max + 1])[0])).normalized()
    V_f_aligned = -V_f if np.dot(omegaHat.vec, V_f.vec) < 0 else V_f

    # R_V_f is the rotor taking the Z axis onto V_f
    R_V_f = (-V_f_aligned * quaternion.z).sqrt()
    # INFOTOCERR << omegaHat << "\n"
    #            << V_f << "\n"
    #            << V_f_aligned << "\n"
    #            << R_V_f * Quaternions::zHat * R_V_f.conjugate() << "\n" << std::endl;

    # Now rotate Instant so that its z axis is aligned with V_f
    Instant.rotate_decomposition_basis(R_V_f)

    # Get the phase of the (2,+/-2) modes after rotation
    i_22 = Instant.index(2, 2)
    i_2m2 = Instant.index(2, -2)
    phase_22 = math.atan2(Instant.data[0, i_22].imag, Instant.data[0, i_22].real)
    phase_2m2 = math.atan2(Instant.data[0, i_2m2].imag, Instant.data[0, i_2m2].real)

    # R_eps is the rotation we will be applying on the right-hand side
    R_eps = R_V_f * (quaternion.quaternion(0, 0, 0, (-(phase_22 - phase_2m2) / 8.0))).exp()

    # Without changing anything else (the direction of V_f or the
    # phase), make sure that the rotating frame's XHat axis is more
    # parallel to the input nHat_t_fid than anti-parallel.
    if np.dot(nHat_t_fid.vec, (R_f0 * R_eps * quaternion.x * R_eps.inverse() * R_f0.inverse()).vec) < 0:
        R_eps = R_eps * ((math.pi / 2.0) * quaternion.z).exp()

    return R_eps


@waveform_alterations
def align_decomposition_frame_to_modes(w, t_fid, nHat_t_fid=quaternion.x, ell_max=None):
    """Fix the attitude of the corotating frame.

    The corotating frame is only defined up to some constant rotor
    R_eps; if R_corot is corotating, then so is R_corot*R_eps.  This
    function uses that freedom to ensure that the frame is aligned
    with the Waveform modes at the fiducial time.  In particular, it
    ensures that the Z axis of the frame in which the decomposition
    is done is along the dominant eigenvector of the <LL> matrix
    (suggested by O'Shaughnessy et al.), and the phase of the (2,2)
    mode is zero.

    If ell_max is None (default), all ell modes are used.

    Parameters
    ----------
    w: WaveformModes
        Object to be aligned
    t_fid: float
        Fiducial time at which the alignment should happen
    nHat_t_fid: quaternion (optional)
        The approximate direction of nHat at t_fid; defaults to x
    ell_max: int
       Highest ell mode to use in computing the <LL> matrix

    """

    # Find the appropriate rotation
    R_eps = get_alignment_of_decomposition_frame_to_modes(w, t_fid, nHat_t_fid, ell_max)

    # Record what happened
    command = "{0}.align_decomposition_frame_to_modes({1}, {2}, {3})  # R_eps={4}"
    w._append_history(command.format(w, t_fid, nHat_t_fid, ell_max, R_eps))

    # Now, apply the rotation
    w = w.rotate_decomposition_basis(R_eps)

    return w


@waveform_alterations
def rotate_physical_system(W, R_phys):
    """Rotate a Waveform in place

    This just rotates the decomposition basis by the inverse of the input
    rotor(s).  See `rotate_decomposition_basis`.

    For more information on the analytical details, see
    http://moble.github.io/spherical_functions/SWSHs.html#rotating-swshs

    """
    W = rotate_decomposition_basis(W, ~R_phys)
    W._append_history(f"{W}.rotate_physical_system({R_phys})")
    return W  # Probably no return, but just in case...


@waveform_alterations
def rotate_decomposition_basis(W, R_basis):
    """Rotate a Waveform in place

    This function takes a Waveform object `W` and either a quaternion
    or array of quaternions `R_basis`.  It applies that rotation to
    the decomposition basis of the modes in the Waveform.  The change
    in basis is also recorded in the Waveform's `frame` data.

    For more information on the analytical details, see
    http://moble.github.io/spherical_functions/SWSHs.html#rotating-swshs

    """
    # This will be used in the jitted functions below to store the
    # Wigner D matrix at each time step
    D = np.empty((sf.WignerD._total_size_D_matrices(W.ell_min, W.ell_max),), dtype=complex)

    if isinstance(R_basis, (list, np.ndarray)) and len(R_basis) == 1:
        R_basis = R_basis[0]

    if isinstance(R_basis, (list, np.ndarray)):
        if isinstance(R_basis, np.ndarray) and R_basis.ndim != 1:
            raise ValueError("Input dimension mismatch.  R_basis.shape={1}".format(R_basis.shape))
        if W.n_times != len(R_basis):
            raise ValueError(
                "Input dimension mismatch.  (W.n_times={}) != (len(R_basis)={})".format(W.n_times, len(R_basis))
            )
        _rotate_decomposition_basis_by_series(W.data, quaternion.as_spinor_array(R_basis), W.ell_min, W.ell_max, D)

        # Update the frame data, using right-multiplication
        if W.frame.size:
            if W.frame.shape[0] == 1:
                # Numpy can't currently multiply one element times an array
                W.frame = np.array([W.frame * R for R in R_basis])
            else:
                W.frame = W.frame * R_basis
        else:
            W.frame = np.copy(R_basis)

    # We can't just use an `else` here because we need to process the
    # case where the input was an iterable of length 1, which we've
    # now changed to just a single quaternion.
    if isinstance(R_basis, np.quaternion):
        sf._Wigner_D_matrices(R_basis.a, R_basis.b, W.ell_min, W.ell_max, D)
        tmp = np.empty((2 * W.ell_max + 1,), dtype=complex)
        _rotate_decomposition_basis_by_constant(W.data, W.ell_min, W.ell_max, D, tmp)

        # Update the frame data, using right-multiplication
        if W.frame.size:
            W.frame = W.frame * R_basis
        else:
            W.frame = np.array([R_basis])

    opts = np.get_printoptions()
    np.set_printoptions(threshold=6)
    W.__history_depth__ -= 1
    W._append_history(f"{W}.rotate_decomposition_basis({R_basis})")
    np.set_printoptions(**opts)

    return W


@jit("void(c16[:,:], i8, i8, c16[:], c16[:])")
def _rotate_decomposition_basis_by_constant(data, ell_min, ell_max, D, tmp):
    """Rotate data by the same rotor at each point in time

    `D` is the Wigner D matrix for all the ell values.

    `tmp` is just a workspace used as temporary storage to hold the
    results for each item of data during the sum.

    """
    for i_t in range(data.shape[0]):
        for ell in range(ell_min, ell_max + 1):
            i_data = ell ** 2 - ell_min ** 2
            i_D = sf._linear_matrix_offset(ell, ell_min)

            for i_m in range(2 * ell + 1):
                tmp[i_m] = 0j
            for i_mp in range(2 * ell + 1):
                for i_m in range(2 * ell + 1):
                    tmp[i_m] += data[i_t, i_data + i_mp] * D[i_D + (2 * ell + 1) * i_mp + i_m]
            for i_m in range(2 * ell + 1):
                data[i_t, i_data + i_m] = tmp[i_m]


@jit("void(c16[:,:], c16[:,:], i8, i8, c16[:])")
def _rotate_decomposition_basis_by_series(data, R_basis, ell_min, ell_max, D):
    """Rotate data by a different rotor at each point in time

    `D` is just a workspace, which holds the Wigner D matrices.
    During the summation, it is also used as temporary storage to hold
    the results for each item of data, where the first row in the
    matrix is overwritten with the new sums.

    """
    for i_t in range(data.shape[0]):
        sf._Wigner_D_matrices(R_basis[i_t, 0], R_basis[i_t, 1], ell_min, ell_max, D)
        for ell in range(ell_min, ell_max + 1):
            i_data = ell ** 2 - ell_min ** 2
            i_D = sf._linear_matrix_offset(ell, ell_min)

            for i_m in range(2 * ell + 1):
                new_data_mp = 0j
                for i_mp in range(2 * ell + 1):
                    new_data_mp += data[i_t, i_data + i_mp] * D[i_D + i_m + (2 * ell + 1) * i_mp]
                D[i_D + i_m] = new_data_mp
            for i_m in range(2 * ell + 1):
                data[i_t, i_data + i_m] = D[i_D + i_m]



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import pytest

import numpy as np
from numpy import *
import quaternion
import spherical_functions as sf

import scri


def pytest_addoption(parser):
    parser.addoption("--run_slow_tests", action="store_true", help="Run all tests, including slow ones")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run_slow_tests"):
        # --run_slow_tests given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --run_slow_tests option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def pytest_runtest_setup(item):
    if "slow" in item.keywords and not item.config.getoption("--run_slow_tests"):
        pytest.skip("Need `--run_slow_tests` command-line argument to run")


def kerr_schild(mass, spin, ell_max=8):
    psi2 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)
    psi1 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)
    psi0 = np.zeros(sf.LM_total_size(0, ell_max), dtype=complex)

    # In the Moreschi-Boyle convention
    psi2[0] = -sf.constant_as_ell_0_mode(mass)
    psi1[2] = -np.sqrt(2) * (3j * spin / 2) * (np.sqrt((8 / 3) * np.pi))
    psi0[6] = 2 * (3 * spin**2 / mass / 2) * (np.sqrt((32 / 15) * np.pi))

    return psi2, psi1, psi0


@pytest.fixture(name="kerr_schild")
def kerr_schild_fixture():
    return kerr_schild()


def constant_waveform(begin=-10.0, end=100.0, n_times=1000, ell_min=2, ell_max=8):
    t = np.linspace(begin, end, num=n_times)
    frame = np.array([quaternion.x for t_i in t])
    lm = np.array([[ell, m] for ell in range(ell_min, ell_max + 1) for m in range(-ell, ell + 1)])
    data = np.empty((t.shape[0], lm.shape[0]), dtype=complex)
    for i, m in enumerate(lm[:, 1]):
        data[:, i] = m - 1j * m
    W = scri.WaveformModes(
        t=t,
        frame=frame,
        data=data,
        ell_min=min(lm[:, 0]),
        ell_max=max(lm[:, 0]),
        history=["# Called from constant_waveform"],
        frameType=scri.Corotating,
        dataType=scri.h,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )
    return W


@pytest.fixture(name="constant_waveform")
def constant_waveform_fixture():
    return constant_waveform()


def linear_waveform(begin=-10.0, end=100.0, n_times=1000, ell_min=2, ell_max=8):
    np.random.seed(hash("linear_waveform") % 4294967294)  # Use mod to get in an acceptable range
    axis = np.quaternion(0.0, *np.random.uniform(-1, 1, size=3)).normalized()
    t = np.linspace(begin, end, num=n_times)
    omega = 2 * np.pi * 4 / (t[-1] - t[0])
    frame = np.array([np.exp(axis * (omega * t_i / 2)) for t_i in t])
    lm = np.array([[ell, m] for ell in range(ell_min, ell_max + 1) for m in range(-ell, ell + 1)])
    data = np.empty((t.shape[0], lm.shape[0]), dtype=complex)
    for i, m in enumerate(lm[:, 1]):
        # N.B.: This form is used in test_linear_interpolation; if you
        # change it here, you must change it there.
        data[:, i] = (m - 1j * m) * t
    W = scri.WaveformModes(
        t=t,
        frame=frame,
        data=data,
        ell_min=min(lm[:, 0]),
        ell_max=max(lm[:, 0]),
        history=["# Called from linear_waveform"],
        frameType=scri.Corotating,
        dataType=scri.h,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )
    return W


@pytest.fixture(name="linear_waveform")
def linear_waveform_fixture():
    return linear_waveform()


def random_waveform(begin=-10.0, end=100.0, n_times=1000, ell_min=None, ell_max=8, dataType=scri.h):
    np.random.seed(hash("random_waveform") % 4294967294)  # Use mod to get in an acceptable range
    spin_weight = scri.SpinWeights[scri.DataType.index(dataType)]
    if ell_min is None:
        ell_min = abs(spin_weight)
    n_modes = ell_max * (ell_max + 2) - ell_min**2 + 1
    t = np.sort(np.random.uniform(begin, end, size=n_times))
    frame = np.array([np.quaternion(*np.random.uniform(-1, 1, 4)).normalized() for t_i in t])
    data = np.random.normal(size=(n_times, n_modes, 2)).view(complex)[:, :, 0]
    W = scri.WaveformModes(
        t=t,
        frame=frame,
        data=data,
        ell_min=ell_min,
        ell_max=ell_max,
        history=["# Called from random_waveform"],
        frameType=scri.Corotating,
        dataType=dataType,
        r_is_scaled_out=True,
        m_is_scaled_out=False,
    )
    return W


@pytest.fixture(name="random_waveform")
def random_waveform_fixture():
    return random_waveform()


def delta_waveform(ell, m, begin=-10.0, end=100.0, n_times=1000, ell_min=2, ell_max=8):
    """WaveformModes with 1 in selected slot and 0 elsewhere"""
    n_modes = ell_max * (ell_max + 2) - ell_min**2 + 1
    t = np.linspace(begin, end, num=n_times)
    data = np.zeros((n_times, n_modes), dtype=complex)
    data[:, sf.LM_index(ell, m, ell_min)] = 1.0 + 0.0j
    W = scri.WaveformModes(
        t=t,
        data=data,  # frame=frame,
        ell_min=ell_min,
        ell_max=ell_max,
        history=["# Called from delta_waveform"],
        frameType=scri.Inertial,
        dataType=scri.psi4,
        r_is_scaled_out=False,
        m_is_scaled_out=True,
    )
    return W


@pytest.fixture(name="delta_waveform")
def delta_waveform_fixture():
    return delta_waveform()


@pytest.fixture
def special_angles():
    return np.arange(-1 * np.pi, 1 * np.pi + 0.1, np.pi / 4.0)


@pytest.fixture
def Rs():
    ones = [0, -1.0, 1.0]
    rs = [np.quaternion(w, x, y, z).normalized() for w in ones for x in ones for y in ones for z in ones][1:]
    np.random.seed(hash("Rs") % 4294967294)  # Use mod to get in an acceptable range
    rs = rs + [np.quaternion(*np.random.uniform(-1, 1, 4)).normalized() for i in range(20)]
    return np.array(rs)


def real_supertranslation(α):
    """Update α in place to represent a real-valued supertranslation."""
    ℓ = 0
    while True:
        for m in range(-ℓ, ℓ+1):
            iₚ, iₘ = sf.LM_index(ℓ, m, 0), sf.LM_index(ℓ, -m, 0)
            if iₚ >= α.size:
                return α
            α[iₚ] = (α[iₚ] + (-1.0)**m * α[iₘ].conj()) / 2
            α[iₘ] = (-1.0)**m * α[iₚ].conj()
        ℓ += 1



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import numpy as np
import quaternion
from numpy import *
import pytest
import scri
from scri.waveform_modes import WaveformModes

from conftest import linear_waveform, constant_waveform, random_waveform


def test_empty_WaveformModes():
    W = scri.WaveformModes()
    assert W.ensure_validity(alter=False)
    assert W.t.shape == (0,)
    assert W.frame.shape == (0,)
    assert W.data.shape == (0, 0)
    assert W.LM.shape == (0, 2)
    assert W.ell_min == 0
    assert W.ell_max == -1
    # assert W.history == ['WaveformModes22 = WaveformModes(...)']
    assert W.frameType == scri.UnknownFrameType
    assert W.dataType == scri.UnknownDataType
    assert not W.r_is_scaled_out  # != True
    assert not W.m_is_scaled_out  # != True


@pytest.mark.xfail
def test_bad_construction():
    # both with and without override_exception_from_invalidity
    assert False


@pytest.mark.xfail
def test_unused_constructor_kwarg():
    assert False


def test_nontrivial_construction(linear_waveform):
    assert linear_waveform is not None  # As long as this was constructed, we're okay


def test_pickling():
    import pickle

    W1 = scri.WaveformModes()
    W1_str = pickle.dumps(W1)
    W2 = pickle.loads(W1_str)
    assert "_WaveformBase__num" in W1.__dict__.keys()
    assert "_WaveformBase__num" in W2.__dict__.keys()
    assert W1._allclose(W2, rtol=0, atol=0, compare_history_beginnings=True)


def test_waveform_properties(linear_waveform):
    W = linear_waveform
    assert W.ensure_validity(alter=False)
    assert W.n_data_sets == len([1 for ell in range(2, 9) for m in range(-ell, ell + 1)])
    assert W.n_times == 1000
    assert W.spin_weight == -2
    assert W.conformal_weight == -1
    assert W.gamma_weight == -1
    assert W.r_scaling == 1
    assert W.m_scaling == 0
    assert W.frame_type_string == "Corotating"
    assert W.data_type_string == "h"
    assert W.data_type_latex == "h"
    assert W.descriptor_string == "rhOverM"
    assert W.ell_min == 2
    assert W.ell_max == 8
    assert W.ells == (2, 8)
    assert np.array_equal(W.LM, np.array([[ell, m] for ell in range(2, 9) for m in range(-ell, ell + 1)]))


def test_indexing(linear_waveform):
    w = linear_waveform
    for i, (ell, m) in enumerate(w.LM):
        assert w.index(ell, m) == i
        assert np.allclose(w.data[:, w.index(ell, m)], (m - 1j * m) * w.t, rtol=0, atol=0)
    assert np.all(w.indices(w.LM) == range(w.n_modes))


def test_empty_slice(linear_waveform):
    W = linear_waveform[:0, :0]
    assert W.ensure_validity(alter=False)
    # Check that these things are copied from linear_waveform
    assert W.history[:1] == ["# Called from linear_waveform"]
    assert W.frameType == scri.Corotating
    assert W.dataType == scri.h
    assert W.r_is_scaled_out  # == True
    assert W.m_is_scaled_out  # == True
    assert W.num != linear_waveform.num
    # Check that these are as in an empty WaveformModes
    assert W.t.shape == (0,)
    assert W.frame.shape == (0,)
    assert W.data.shape == (0, 0)
    assert W.ell_min == 0
    assert W.ell_max == -1


def test_empty_mode_slice(linear_waveform):
    W = linear_waveform[:, :0]
    assert W.ensure_validity(alter=False)
    # Check that these things are copied from linear_waveform
    assert np.all(W.t == linear_waveform.t)
    assert np.all(W.frame == linear_waveform.frame)
    # Check that these are empty
    assert W.data.size == 0
    assert W.LM.size == 0
    assert W.ell_min == 0
    assert W.ell_max == -1
    assert W.history[:1] == ["# Called from linear_waveform"]
    assert W.frameType == linear_waveform.frameType
    assert W.dataType == linear_waveform.dataType
    assert W.r_is_scaled_out == linear_waveform.r_is_scaled_out
    assert W.m_is_scaled_out == linear_waveform.m_is_scaled_out
    assert W.num != linear_waveform.num  # Check that this is valid, but not the same


def test_time_slice(linear_waveform):
    W = linear_waveform[10:50]
    assert W.ensure_validity(alter=False)
    assert np.all(W.t == linear_waveform.t[10:50])
    assert np.all(W.frame == linear_waveform.frame[10:50])
    assert np.all(W.data == linear_waveform.data[10:50])
    assert np.all(W.LM == linear_waveform.LM)
    assert W.ell_min == linear_waveform.ell_min
    assert W.ell_max == linear_waveform.ell_max
    assert W.history[:1] == ["# Called from linear_waveform"]
    assert W.frameType == linear_waveform.frameType
    assert W.dataType == linear_waveform.dataType
    assert W.r_is_scaled_out == linear_waveform.r_is_scaled_out
    assert W.m_is_scaled_out == linear_waveform.m_is_scaled_out
    assert isinstance(W.num, int)
    assert W.num != linear_waveform.num


def test_time_and_mode_slice(linear_waveform):
    W = linear_waveform[10:50, :5]
    assert W.ensure_validity(alter=False)
    assert W.history[:1] == ["# Called from linear_waveform"]
    assert W.frameType == scri.Corotating
    assert W.dataType == scri.h
    assert W.r_is_scaled_out  # == True
    assert W.m_is_scaled_out  # == True
    assert W.num != linear_waveform.num
    assert np.all(W.t == linear_waveform.t[10:50])
    assert np.all(W.frame == linear_waveform.frame[10:50])
    assert np.all(
        W.LM == np.array([[ell, m] for ell in range(linear_waveform.ell_min, 5) for m in range(-ell, ell + 1)])
    )
    assert np.all(W.data == linear_waveform.data[10:50, :21])


def test_norms(linear_waveform):
    W = linear_waveform
    W.data[0, :] = 6.0 * W.data[-1, :]
    W.data[10, :] = 5.0 * W.data[-1, :]
    assert W.ensure_validity(alter=False)
    assert np.allclose(W.norm(), np.sum(np.abs(W.data) ** 2, axis=-1), rtol=1.0e-15)
    assert np.allclose(W.norm(take_sqrt=True), np.sqrt(np.sum(np.abs(W.data) ** 2, axis=-1)), rtol=1.0e-15)
    assert np.allclose(
        W.norm(indices=slice(W.data.shape[0] // 4, None, None)),
        np.sum(np.abs(W.data[(W.data.shape[0] // 4) :]) ** 2, axis=-1),
        rtol=1.0e-15,
    )
    assert np.allclose(
        W.norm(indices=slice(W.data.shape[0] // 4, None, None), take_sqrt=True),
        np.sqrt(np.sum(np.abs(W.data[(W.data.shape[0] // 4) :]) ** 2, axis=-1)),
        rtol=1.0e-15,
    )
    assert W.max_norm_index() == W.data.shape[0] - 1
    assert W.max_norm_index(0) == 0
    assert W.max_norm_index(1) == 0
    assert W.max_norm_index(W.data.shape[0]) == 10
    assert W.max_norm_time() == W.t[-1]
    assert W.max_norm_time(0) == W.t[0]
    assert W.max_norm_time(1) == W.t[0]
    assert W.max_norm_time(W.data.shape[0]) == W.t[10]


@pytest.mark.parametrize("W", [linear_waveform, constant_waveform])
def test_interpolation(W):
    interpolation_precision = 4.5e-16

    # Interpolation onto self.t is the identity (except for `num`)
    W_in = W()
    t_in = W_in.t
    t_out = np.copy(t_in)
    W_out = W_in.interpolate(t_out)
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    assert W_out.history[:2] == W_in.history[:2]
    assert W_out.frameType == W_in.frameType
    assert W_out.dataType == W_in.dataType
    assert W_out.r_is_scaled_out == W_in.r_is_scaled_out
    assert W_out.m_is_scaled_out == W_in.m_is_scaled_out
    assert W_out.num != W_in.num
    assert np.all(W_out.t == W_in.t)
    assert np.all(W_out.frame == W_in.frame)
    assert np.all(W_out.LM == W_in.LM)
    assert np.allclose(W_out.data, W_in.data, rtol=4.5e-16)

    # Interpolation onto another time series should reshape
    W_in = W()
    t_in = W_in.t
    t_out = (t_in[:-1] + t_in[1:]) / 2.0
    W_out = W_in.interpolate(t_out)
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    assert W_out.history[:1] == [f"# Called from {W.__name__}"]
    assert W_out.frameType == scri.Corotating
    assert W_out.dataType == scri.h
    assert W_out.r_is_scaled_out  # == True
    assert W_out.m_is_scaled_out  # == True
    assert isinstance(W_out.num, int)
    assert W_out.num != W_in.num
    assert np.all(W_out.t == t_out)
    assert np.all(W_out.frame == quaternion.squad(W_in.frame, W_in.t, W_out.t))
    assert np.all(W_out.LM == W_in.LM)
    assert W_out.data.shape == (len(t_out), W_in.n_data_sets)


def test_constant_interpolation(constant_waveform):
    interpolation_precision = 4.5e-16
    # Interpolation of constant_waveform is ~identity
    W_in = constant_waveform
    t_in = W_in.t
    t_out = (t_in[:-1] + t_in[1:]) / 2.0
    W_out = W_in.interpolate(t_out)
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    assert W_out.history[:1] == ["# Called from constant_waveform"]
    assert W_out.frameType == scri.Corotating
    assert W_out.dataType == scri.h
    assert W_out.r_is_scaled_out  # == True
    assert W_out.m_is_scaled_out  # == True
    assert W_out.num != W_in.num
    assert np.all(W_out.t == t_out)
    assert np.all(W_out.frame == np.array([W_in.frame[0],] * len(t_out)))
    assert np.all(W_out.LM == W_in.LM)
    assert np.allclose(W_out.data, np.array([W_in.data[0, :],] * len(t_out)), rtol=interpolation_precision)


def test_linear_interpolation(linear_waveform):
    interpolation_precision = 4.5e-16
    # Interpolation of constant_waveform is ~identity
    W_in = linear_waveform
    t_in = W_in.t
    t_out = (t_in[:-1] + t_in[1:]) / 2.0
    W_out = W_in.interpolate(t_out)
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    assert W_out.history[:1] == ["# Called from linear_waveform"]
    assert W_out.frameType == scri.Corotating
    assert W_out.dataType == scri.h
    assert W_out.r_is_scaled_out  # == True
    assert W_out.m_is_scaled_out  # == True
    assert W_out.num != W_in.num
    assert np.all(W_out.t == t_out)
    squad = quaternion.squad(W_in.frame, W_in.t, t_out)
    assert np.all(np.abs(W_out.frame - squad) < interpolation_precision * np.abs(W_out.frame + squad))
    assert np.all(W_out.LM == W_in.LM)
    data = np.empty((t_out.shape[0], W_in.LM.shape[0]), dtype=complex)
    for i, m in enumerate(W_in.LM[:, 1]):
        # N.B.: This depends on the construction of linear_waveform;
        # only change here if it is changed there.
        data[:, i] = (m - 1j * m) * t_out
    assert np.allclose(W_out.data, data, rtol=interpolation_precision)


def test_involution_properties(random_waveform):
    """Ensure that involutions truly are involute

    The objective here is to apply the involutions twice, and ensure that we get the same thing out.

    """
    for direction in ["x_", "y_", "z_", ""]:
        w_in = WaveformModes(random_waveform)
        w_out = getattr(getattr(w_in, f"{direction}parity_conjugate"), f"{direction}parity_conjugate")
        assert w_in._allclose(w_out, rtol=0.0, atol=0.0, compare_history_beginnings=True)


def test_idempotents(random_waveform):
    """Ensure that (anti)symmetrization composed with itself is itself

    Taking the symmetric part of the symmetric part should be the same as just the original symmetric part.  The same
    goes for anti-symmetric parts.

    """
    for direction in ["x_", "y_", "z_", ""]:
        w_in = getattr(random_waveform, f"{direction}parity_symmetric_part")
        w_out = getattr(w_in, f"{direction}parity_symmetric_part")
        assert w_in._allclose(w_out, rtol=0.0, atol=0.0, compare_history_beginnings=True)


def test_null_compositions(random_waveform):
    """Ensure that symmetric parts of antisymmetric parts are zero, and vice versa

    Both the `frame` and `data` members should be zeroed out by such operations.

    """
    w_in = WaveformModes(random_waveform)
    w_in.data = np.zeros_like(w_in.data)
    w_in.frame = np.zeros_like(w_in.frame)
    for direction in ["x_", "y_", "z_", ""]:
        w_out = getattr(
            getattr(random_waveform, f"{direction}parity_symmetric_part"), f"{direction}parity_antisymmetric_part"
        )
        assert w_in._allclose(w_out, rtol=0.0, atol=0.0)
    for direction in ["x_", "y_", "z_", ""]:
        w_out = getattr(
            getattr(random_waveform, f"{direction}parity_antisymmetric_part"), f"{direction}parity_symmetric_part"
        )
        assert w_in._allclose(w_out, rtol=0.0, atol=0.0)


def test_parity_violation_measures(random_waveform):
    """Apply parity-part operators, then measure parity violation

    The parity-violation of the parity-symmetric part should be zero.  The parity-violation of the original waveform
    should equal the norm of the parity-antisymmetric part (or 1 for the normalized violation).

    """
    w_in = WaveformModes(random_waveform)
    zeros = np.zeros_like(w_in.data[:, 0])
    ones = np.ones_like(w_in.data[:, 0])
    for direction in ["x_", "y_", "z_", ""]:
        w_out = getattr(random_waveform, f"{direction}parity_symmetric_part")
        assert np.allclose(getattr(w_out, f"{direction}parity_violation_squared"), zeros, atol=1e-15)
        assert np.allclose(getattr(w_out, f"{direction}parity_violation_normalized"), zeros, atol=1e-15)
    for direction in ["x_", "y_", "z_", ""]:
        w_out = getattr(random_waveform, f"{direction}parity_antisymmetric_part")
        assert np.allclose(getattr(w_in, f"{direction}parity_violation_squared"), w_out.norm(), atol=0.0, rtol=1e-15)
        assert np.allclose(
            getattr(w_in, f"{direction}parity_violation_normalized"),
            np.sqrt(w_out.norm() / w_in.norm()),
            atol=0.0,
            rtol=1e-15,
        )
        assert np.allclose(getattr(w_out, f"{direction}parity_violation_normalized"), ones, atol=0.0, rtol=1e-15)


@pytest.mark.xfail
def test_involutions():
    assert False


def test_SI_units(linear_waveform):
    units_precision = 4.5e-16
    TotalMassInSolarMasses = 1.1
    DistanceInMegaparsecs = 3.7
    SpeedOfLight = 299792458  # Units of meters / seconds
    SolarMassInSeconds = 4.92549094916e-06  # Units of seconds
    MassInSeconds = TotalMassInSolarMasses * SolarMassInSeconds
    OneMegaparsec = 3.0856775814913672789e22  # Units of meters
    DistanceInMeters = DistanceInMegaparsecs * OneMegaparsec
    W_in = linear_waveform
    W_out = W_in.SI_units(TotalMassInSolarMasses, DistanceInMegaparsecs)
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    assert W_out.history[:1] == ["# Called from linear_waveform"]
    assert W_out.frameType == scri.Corotating
    assert W_out.dataType == scri.h
    assert not W_out.r_is_scaled_out  # != True
    assert not W_out.m_is_scaled_out  # != True
    assert W_out.num != W_in.num
    assert np.allclose(W_out.t, W_in.t * MassInSeconds, rtol=units_precision)
    assert np.allclose(W_out.data, W_in.data * MassInSeconds * SpeedOfLight / DistanceInMeters, rtol=units_precision)



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the AsymptoticBondData
### class.  In particular, they assume that the first argument, `self` is an instance of
### AsymptoticBondData.  They should probably not be used outside of that class.


def bondi_constraints(self, lhs=True, rhs=True):
    """Compute Bondi-gauge constraint equations

    Bondi gauge establishes some relations that the data must satisfy:

        ψ̇₀ = ðψ₁ + 3 σ ψ₂

        ψ̇₁ = ðψ₂ + 2 σ ψ₃

        ψ̇₂ = ðψ₃ + 1 σ ψ₄

        ψ₃ = -∂ðσ̄/∂u

        ψ₄ = -∂²σ̄/∂u²

        Im[ψ₂] = -Im[ð²σ̄ + σ ∂σ̄/∂u]

    This function returns a 6-tuple of 2-tuples, corresponding to these 6 equations
    and their left- and right-hand sides.

    """
    return (
        self.bianchi_0(lhs, rhs),
        self.bianchi_1(lhs, rhs),
        self.bianchi_2(lhs, rhs),
        self.constraint_3(lhs, rhs),
        self.constraint_4(lhs, rhs),
        self.constraint_mass_aspect(lhs, rhs),
    )


@property
def bondi_violations(self):
    """Compute violations of Bondi-gauge constraints

    Bondi gauge establishes some relations that the data must satisfy:

        ψ̇₀ = ðψ₁ + 3 σ ψ₂

        ψ̇₁ = ðψ₂ + 2 σ ψ₃

        ψ̇₂ = ðψ₃ + 1 σ ψ₄

        ψ₃ = -∂ðσ̄/∂u

        ψ₄ = -∂²σ̄/∂u²

        Im[ψ₂] = -Im[ð²σ̄ + σ ∂σ̄/∂u]

    This function returns a tuple of 6 arrays, corresponding to these 6 equations,
    in which the right-hand side is subtracted from the left-hand side.  No norms
    are taken.

    """
    constraints = self.bondi_constraints(True, True)
    return [lhs - rhs for (lhs, rhs) in constraints]


@property
def bondi_violation_norms(self):
    """Compute norms of violations of Bondi-gauge conditions

    Bondi gauge establishes some relations that the data must satisfy:

        ψ̇₀ = ðψ₁ + 3 σ ψ₂

        ψ̇₁ = ðψ₂ + 2 σ ψ₃

        ψ̇₂ = ðψ₃ + 1 σ ψ₄

        ψ₃ = -∂ðσ̄/∂u

        ψ₄ = -∂²σ̄/∂u²

        Im[ψ₂] = -Im[ð²σ̄ + σ ∂σ̄/∂u]

    This function returns a tuple of 6 arrays, corresponding to the norms of these
    6 equations, in which the right-hand side is subtracted from the left-hand
    side, and then the squared magnitude of that result is integrated over the
    sphere.  No integration is performed over time.

    """
    violations = self.bondi_violations
    return [v.norm() for v in violations]


def bianchi_0(self, lhs=True, rhs=True):
    """Return the left- and/or right-hand sides of the Psi0 component of the Bianchi identity

    In Bondi coordinates, the Bianchi identities simplify, resulting in this
    expression (among others) for the time derivative of ψ₀:

        ψ̇₀ = ðψ₁ + 3 σ ψ₂

    Parameters
    ==========
    lhs: bool [defaults to True]
        If True, return the left-hand side of the equation above
    rhs: bool [defaults to True]
        If True, return the right-hand side of the equation above

    If both parameters are True, a tuple with elements (lhs_value, rhs_value) is
    returned; otherwise just the requested value is returned.

    """
    if lhs:
        lhs_value = self.psi0.dot
    if rhs:
        rhs_value = self.psi1.eth_GHP + 3 * self.sigma * self.psi2
    if lhs and rhs:
        return (lhs_value, rhs_value)
    elif lhs:
        return lhs_value
    elif rhs:
        return rhs_value


def bianchi_1(self, lhs=True, rhs=True):
    """Return the left- and/or right-hand sides of the Psi1 component of the Bianchi identity

    In Bondi coordinates, the Bianchi identities simplify, resulting in this
    expression (among others) for the time derivative of ψ₁:

        ψ̇₁ = ðψ₂ + 2 σ ψ₃

    Parameters
    ==========
    lhs: bool [defaults to True]
        If True, return the left-hand side of the equation above
    rhs: bool [defaults to True]
        If True, return the right-hand side of the equation above

    If both parameters are True, a tuple with elements (lhs_value, rhs_value) is
    returned; otherwise just the requested value is returned.

    """
    if lhs:
        lhs_value = self.psi1.dot
    if rhs:
        rhs_value = self.psi2.eth_GHP + 2 * self.sigma * self.psi3
    if lhs and rhs:
        return (lhs_value, rhs_value)
    elif lhs:
        return lhs_value
    elif rhs:
        return rhs_value


def bianchi_2(self, lhs=True, rhs=True):
    """Return the left- and/or right-hand sides of the Psi2 component of the Bianchi identity

    In Bondi coordinates, the Bianchi identities simplify, resulting in this
    expression (among others) for the time derivative of ψ₂:

        ψ̇₂ = ðψ₃ + σ ψ₄

    Parameters
    ==========
    lhs: bool [defaults to True]
        If True, return the left-hand side of the equation above
    rhs: bool [defaults to True]
        If True, return the right-hand side of the equation above

    If both parameters are True, a tuple with elements (lhs_value, rhs_value) is
    returned; otherwise just the requested value is returned.

    """
    if lhs:
        lhs_value = self.psi2.dot
    if rhs:
        rhs_value = self.psi3.eth_GHP + self.sigma * self.psi4
    if lhs and rhs:
        return (lhs_value, rhs_value)
    elif lhs:
        return lhs_value
    elif rhs:
        return rhs_value


def constraint_3(self, lhs=True, rhs=True):
    """Return the left- and/or right-hand sides of the Psi3 expression in Bondi gauge

    In Bondi coordinates, the value of ψ₃ is given by a time derivative and an
    angular derivative of the (conjugate) shear:

        ψ₃ = -∂ðσ̄/∂u

    Parameters
    ==========
    lhs: bool [defaults to True]
        If True, return the left-hand side of the equation above
    rhs: bool [defaults to True]
        If True, return the right-hand side of the equation above

    If both parameters are True, a tuple with elements (lhs_value, rhs_value) is
    returned; otherwise just the requested value is returned.

    """
    if lhs:
        lhs_value = self.psi3
    if rhs:
        rhs_value = -self.sigma.bar.dot.eth_GHP
    if lhs and rhs:
        return (lhs_value, rhs_value)
    elif lhs:
        return lhs_value
    elif rhs:
        return rhs_value


def constraint_4(self, lhs=True, rhs=True):
    """Return the left- and/or right-hand sides of the Psi4 expression in Bondi gauge

    In Bondi coordinates, the value of ψ₄ is given by the second time derivative
    of the (conjugate) shear:

        ψ₄ = -∂²σ̄/∂u²

    Parameters
    ==========
    lhs: bool [defaults to True]
        If True, return the left-hand side of the equation above
    rhs: bool [defaults to True]
        If True, return the right-hand side of the equation above

    If both parameters are True, a tuple with elements (lhs_value, rhs_value) is
    returned; otherwise just the requested value is returned.

    """
    if lhs:
        lhs_value = self.psi4
    if rhs:
        rhs_value = -self.sigma.bar.ddot
    if lhs and rhs:
        return (lhs_value, rhs_value)
    elif lhs:
        return lhs_value
    elif rhs:
        return rhs_value


def constraint_mass_aspect(self, lhs=True, rhs=True):
    """Return the left- and/or right-hand sides of the mass-aspect reality condition in Bondi gauge

    In Bondi coordinates, the Bondi mass aspect is always real, resulting in this
    relationship:

        Im[ψ₂] = -Im[ð²σ̄ + σ ∂σ̄/∂u]

    Parameters
    ==========
    lhs: bool [defaults to True]
        If True, return the left-hand side of the equation above
    rhs: bool [defaults to True]
        If True, return the right-hand side of the equation above

    If both parameters are True, a tuple with elements (lhs_value, rhs_value) is
    returned; otherwise just the requested value is returned.

    """
    import numpy as np

    if lhs:
        lhs_value = np.imag(self.psi2)
    if rhs:
        rhs_value = -np.imag(self.sigma.bar.eth_GHP.eth_GHP + self.sigma * self.sigma.bar.dot)
    if lhs and rhs:
        return (lhs_value, rhs_value)
    elif lhs:
        return lhs_value
    elif rhs:
        return rhs_value



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from .waveform_base import WaveformBase, waveform_alterations

import warnings
import numpy as np
import spherical_functions as sf
from . import *


class WaveformModes(WaveformBase):
    """Object containing time, frame, and data, along with related information

    This object collects all the data needed to manipulate time-series of spin-weighted spherical-harmonic modes of
    gravitational waves, as well as a few useful informational members.  Various methods exist to manipulate the data
    and extract important information.

    As much as possible, completeness of modes is enforced -- meaning that if any (ell,m) mode with a given ell is
    included in the data, all modes with abs(m)<=ell must also be included.  The data are stored as complex numbers,
    and assumed to be in standard order corresponding to

        [f(ell,m) for ell in range(ell_min, ell_max+1) for m in range(-ell,ell+1)]

    There is no automated check that can be done to assure that the order is correct.  (Precession breaks all
    symmetries, for example.)  The only check that is done ensures that the input data array has the correct
    dimensions.  It is up to the user to ensure that the order is correct.

    Attributes
    ----------
    t : float array
        Time steps corresponding to other data
    frame : quaternion array
        Rotors taking static basis onto decomposition basis
    data : 2-d complex array
        Mode values of the spin-weighted spherical-harmonic decomposition. The array is assumed to have a first
        dimension equal to the length of `t`, and a second dimension equal to the number of modes as deduced from the
        values of ell_min and ell_max below.  As noted above, a very particular order is assumed for the data.
        Input is treated as a reference, so in-place transformations may modify your original array. To avoid this
        behavior, pass a copy of your array.
    ell_min : int
        Smallest ell value present in `data`.
    ell_max : int
        Largest ell value present in `data`.
    LM : int array (read only)
        Array of (ell,m) values corresponding to the `data` member.  This is automatically constructed based on the
        values of ell_min and ell_max, and cannot be reassigned.
    history : list of strings
        As far as possible, all functions applied to the object are recorded in the `history` variable.  In fact,
        the object should almost be able to be recreated using the commands in the history list. Commands taking
        large arrays, however, are shortened -- so the data will not be entirely reconstructable.
    frameType : int
        Index corresponding to `scri.FrameType` appropriate for `data`.
    dataType : int
        Index corresponding to `scri.DataType` appropriate for `data`.
    r_is_scaled_out : bool
        True if the `data` have been multiplied by the appropriate power of radius so that the asymptotic value can
        be finite and nonzero.
    m_is_scaled_out : bool
        True if the `data` have been scaled by the appropriate value of the total mass so that they are dimensionless.
    num : int (read only)
        Automatically assigned number of this object.  The constructor of this type keeps count of the number of
        objects it has created, to assign each object a more-or-less unique ID for use in the history strings.  This
        counter is reset at the beginning of each python session.  Subclasses should automatically have a different
        counter.

    Indexing
    --------
    WaveformMode objects can be indexed much like a numpy array, where the first dimension gives the time indices,
    and the second gives the mode indices. This will return another WaveformMode object containing slices of the
    original data.

    It is important to note, however, that as with numpy array slices, slicing a WaveformMode object will not
    typically copy the original data; the result will simply be a view into the data.  This means that changing the
    data in the slice can change the data in the original.  If you want to make a copy, you should probably use the
    copy constructor: `W2 = WaveformMode(W1)`. It is also possible to use the standard copy.deepcopy method.

    Also note that the first slice dimension corresponds to the indices of the time data, but the second dimension
    does NOT correspond to indices.  Instead, because this object tries to enforce completeness of the mode data,
    you can only slice a WaveformBase with respect to modes.  If you really want just some of the data, you'll need
    to extract it as a plain array, and then just slice as usual.

    For example,

    >>> W  = WaveformModes()
    >>> W[10:-20]

    will give all modes in the data, but only at times starting with the 10th time step, and ending one before the
    -20th time step.  Meanwhile,

    >>> W[10:-20,2]

    will give the same range of times, but only the ell=2 modes -- which includes (ell,m) modes (2,-2) through (2, 2).
    Similarly,

    >>> W[10:-20,2:5]

    will return the same range of times, along with the ell=2,3,4 modes. Note the lack of ell=5 mode, for consistency
    with python's usual slice syntax.

    >>> W[:,:0]

    will return all time steps, along with all `frame` data, but the `lm` data and `data` will be empty (because the
    `:0` term selects everything before the 0th element).  Similarly,

    >>> W[:0,:0]

    is empty of all the numerical data.



    """

    def __init__(self, *args, **kwargs):
        """Initializer for WaveformModes object

        This initializer is primarily a wrapper around the WaveformBase initializer.  See the docstring of
        WaveformBase for more details.  The only difference in calling is that this takes two additional keyword
        parameters:

        Keyword parameters
        ------------------
        ell_min : int, defaults to 0
        ell_max : int, defaults to -1

        """
        if len(args) == 1 and isinstance(args[0], type(self)):
            other = args[0]
            # Do not directly access __ell_min, __ell_max, or __LM outside of this initializer function; use ell_min,
            #  ell_max, or LM instead
            self.__ell_min = other.__ell_min
            self.__ell_max = other.__ell_max
            self.__LM = np.copy(other.__LM)
        else:
            # Do not directly access __ell_min, __ell_max, or __LM outside of this initializer function; use ell_min,
            # ell_max, or LM instead
            self.__ell_min = kwargs.pop("ell_min", 0)
            self.__ell_max = kwargs.pop("ell_max", -1)
            self.__LM = sf.LM_range(self.__ell_min, self.__ell_max)
        super().__init__(*args, **kwargs)

    @classmethod
    def from_sxs(cls, w_sxs, override_exception_from_invalidity=False):
        """Construct this object from an `sxs.WaveformModes` object

        Note that the resulting object will likely contain references to the same
        underlying data contained in the original object; modifying one will modify the
        other.  You can make a copy of the result — using code like
        `WaveformModes.from_sxs(w_sxs).copy()` — to obtain separate data.

        """
        import quaternion
        constructor_statement = (
            f"WaveformModes.from_sxs({w_sxs}, "
            f"override_exception_from_invalidity={override_exception_from_invalidity})"
        )

        try:
            frameType = [n.lower() for n in FrameNames].index(w_sxs.frame_type.lower())
        except ValueError:
            frameType = 0

        try:
            dataType = [n.lower() for n in DataNames].index(w_sxs.data_type.lower())
        except ValueError:
            dataType = 0

        history = w_sxs._metadata.get("history", [])
        if isinstance(history, str):
            history = history.split("\n")

        kwargs = dict(
            t=w_sxs.t,
            #frame=,  # see below
            data=w_sxs.data,
            history=history,
            version_hist=w_sxs._metadata.get("version_hist", []),
            frameType=frameType,
            dataType=dataType,
            r_is_scaled_out=w_sxs._metadata.get("r_is_scaled_out", True),
            m_is_scaled_out=w_sxs._metadata.get("m_is_scaled_out", True),
            override_exception_from_invalidity=override_exception_from_invalidity,
            constructor_statement=constructor_statement,
            ell_min=w_sxs.ell_min,
            ell_max=w_sxs.ell_max,
        )

        frame = w_sxs.frame.ndarray
        if np.array_equal(frame, [[1.,0,0,0]]):
            pass  # The default will handle itself
        elif frame.shape[0] == 1:
            kwargs["frame"] = quaternion.as_quat_array(frame[0, :])
        elif frame.shape[0] == w_sxs.n_times:
            kwargs["frame"] = quaternion.as_quat_array(frame)
        else:
            raise ValueError(
                f"Frame size ({frame.size}) should be 1 or "
                f"equal to the number of time steps ({self.n_times})"
            )

        return cls(**kwargs)

    @property
    def to_sxs(self):
        """Convert this object to an `sxs.WaveformModes` object

        Note that the resulting object will likely contain references to the same
        underlying data contained in the original object; modifying one will modify the
        other.  You can make a copy of this object *before* calling this function —
        using code like `w.copy().to_sxs` — to obtain separate data.

        """
        import sxs
        import quaternionic
        import quaternion
        from .extrapolation import extrapolate

        # All of these will be stored in the `_metadata` member of the resulting WaveformModes
        # object; most of these will also be accessible directly as attributes.
        kwargs = dict(
            time=self.t,
            time_axis=0,
            modes_axis=1,
            #frame=,  # see below
            spin_weight=self.spin_weight,
            data_type=self.data_type_string.lower(),
            frame_type=self.frame_type_string.lower(),
            history=self.history,
            version_hist=self.version_hist,
            r_is_scaled_out=self.r_is_scaled_out,
            m_is_scaled_out=self.m_is_scaled_out,
            ell_min=self.ell_min,
            ell_max=self.ell_max,
        )

        # If self.frame.size==0, we just don't pass any argument
        if self.frame.size == 1:
            kwargs["frame"] = quaternionic.array([quaternion.as_float_array(self.frame)])
        elif self.frame.size == self.n_times:
            kwargs["frame"] = quaternionic.array(quaternion.as_float_array(self.frame))
        elif self.frame.size > 0:
            raise ValueError(
                f"Frame size ({self.frame.size}) should be 0, 1, or "
                f"equal to the number of time steps ({self.n_times})"
            )

        w = sxs.WaveformModes(self.data, **kwargs)

        # Special cases for extrapolate_coord_radii and translation/boost
        if hasattr(self, "extrapolate_coord_radii"):
            w.register_modification(
                extrapolate,
                CoordRadii=list(self.extrapolate_coord_radii),
            )
        if hasattr(self, "space_translation") or hasattr(self, "boost_velocity"):
            w.register_modification(
                self.transform,
                space_translation=list(getattr(self, "space_translation", [0., 0., 0.])),
                boost_velocity=list(getattr(self, "boost_velocity", [0., 0., 0.])),
            )

        return w

    @waveform_alterations
    def ensure_validity(self, alter=True, assertions=False):
        """Try to ensure that the `WaveformModes` object is valid

        See `WaveformBase.ensure_validity` for the basic tests.  This function also includes tests that `data` is
        complex, and consistent with the ell_min and ell_max values.

        """
        import numbers

        errors = []
        alterations = []

        if assertions:
            from .waveform_base import test_with_assertions

            test = test_with_assertions
        else:
            from .waveform_base import test_without_assertions

            test = test_without_assertions

        # We first need to check that the ell values make sense,
        # because we'll use these below
        test(
            errors,
            isinstance(self.__ell_min, numbers.Integral),
            "isinstance(self.__ell_min, numbers.Integral) # type(self.__ell_min)={}".format(type(self.__ell_min)),
        )
        test(
            errors,
            isinstance(self.__ell_max, numbers.Integral),
            "isinstance(self.__ell_max, numbers.Integral) # type(self.__ell_max)={}".format(type(self.__ell_max)),
        )
        test(errors, self.__ell_min >= 0, f"self.__ell_min>=0 # {self.__ell_min}")
        test(
            errors,
            self.__ell_max >= self.__ell_min - 1,
            "self.__ell_max>=self.__ell_min-1 # self.__ell_max={}; self.__ell_min-1={}".format(
                self.__ell_max, self.__ell_min - 1
            ),
        )
        if alter and not np.array_equal(self.__LM, sf.LM_range(self.ell_min, self.ell_max)):
            self.__LM = sf.LM_range(self.ell_min, self.ell_max)
            alterations += [
                "{}._{}__LM = sf.LM_range({}, {})".format(self, type(self).__name__, self.ell_min, self.ell_max)
            ]
        test(
            errors,
            np.array_equal(self.__LM, sf.LM_range(self.ell_min, self.ell_max)),
            "np.array_equal(self.__LM, sf.LM_range(self.ell_min, self.ell_max))",
        )

        test(
            errors,
            self.data.dtype == np.dtype(complex),
            f"self.data.dtype == np.dtype(complex) # self.data.dtype={self.data.dtype}",
        )
        test(errors, self.data.ndim >= 2, f"self.data.ndim >= 2 # self.data.ndim={self.data.ndim}")
        test(
            errors,
            self.data.shape[1] == self.__LM.shape[0],
            "self.data.shape[1]==self.__LM.shape[0] "
            "# self.data.shape={}; self.__LM.shape[0]={}".format(self.data.shape[1], self.__LM.shape[0]),
        )

        if alterations:
            self._append_history(alterations)
            print("The following alterations were made:\n\t" + "\n\t".join(alterations))
        if errors:
            print("The following conditions were found to be incorrectly False:\n\t" + "\n\t".join(errors))
            return False

        # Call the base class's version
        super().ensure_validity(alter, assertions)

        self.__history_depth__ -= 1
        self._append_history("WaveformModes.ensure_validity" + f"({self}, alter={alter}, assertions={assertions})")

        return True

    # Mode information
    @property
    def n_modes(self):
        return self.data.shape[1]

    @property
    def ell_min(self):
        return self.__ell_min

    @ell_min.setter
    def ell_min(self, new_ell_min):
        self.__ell_min = new_ell_min
        self.__LM = sf.LM_range(self.ell_min, self.ell_max)
        if self.n_modes != self.__LM.shape[0]:
            warning = (
                f"\nWaveform's data.shape={self.data.shape} does not agree with "
                + f"(ell_min,ell_max)=({self.ell_min},{self.ell_max}).\n"
                + "Hopefully you are about to reset `data`.  To suppress this warning,\n"
                + "reset `data` before resetting ell_min and/or ell_max."
            )
            warnings.warn(warning)

    @property
    def ell_max(self):
        return self.__ell_max

    @ell_max.setter
    def ell_max(self, new_ell_max):
        self.__ell_max = new_ell_max
        self.__LM = sf.LM_range(self.ell_min, self.ell_max)
        if self.n_modes != self.__LM.shape[0]:
            warning = (
                f"\nWaveform's data.shape={self.data.shape} does not agree with "
                + f"(ell_min,ell_max)=({self.ell_min},{self.ell_max}).\n"
                + "Hopefully you are about to reset `data`.  To suppress this warning,\n"
                + "reset `data` before resetting ell_min and/or ell_max."
            )
            warnings.warn(warning)

    @property
    def ells(self):
        """Return self.ell_min,self.ell_max"""
        return self.__ell_min, self.__ell_max

    @ells.setter
    def ells(self, new_ells):
        """Setting both at once can be necessary when changing the shape of `data`"""
        self.__ell_min = new_ells[0]
        self.__ell_max = new_ells[1]
        self.__LM = sf.LM_range(self.ell_min, self.ell_max)
        if self.n_modes != self.__LM.shape[0]:
            warning = (
                f"\nWaveform's data.shape={self.data.shape} does not agree with "
                + f"(ell_min,ell_max)=({self.ell_min},{self.ell_max}).\n"
                + "Hopefully you are about to reset `data`.  To avoid this warning,\n"
                + "reset `data` before resetting ell_min and/or ell_max."
            )
            warnings.warn(warning)

    @property
    def LM(self):
        """Array of (ell,m) values in Waveform data

        This array is just a flat array of `[ell,m]` pairs.  It is
        automatically recomputed each time `ell_min` or `ell_max` is
        changed.  Specifically, it is

            np.array([[ell,m]
                      for ell in range(self.ell_min,self.ell_max+1)
                      for m in range(-ell,ell+1)])

        """
        return self.__LM

    @LM.setter
    def LM(self, newLM):
        raise AttributeError("Can't set LM data.  This can only be controlled through the ell_min and ell_max members.")

    def index(self, ell, m):
        """Index of given (ell,m) mode in the data

        Parameters
        ----------
        ell: int
        m: int

        Returns
        -------
        idx: int
            Index such that self.LM[idx] is [ell, m]

        """
        return sf.LM_index(ell, m, self.ell_min)

    def indices(self, ell_m):
        """Indices of given (ell,m) modes in the data

        Parameters
        ----------
        ell_m: Nx2 array-like, int

        Returns
        -------
        idcs: array of int
            Index such that self.LM[idcs] is `ell_m`

        """
        ell_m = np.array(ell_m, copy=False)
        if not (ell_m.dtype == int and ell_m.shape[1] == 2):
            raise ValueError("Input `ell_m` should be an Nx2 sequence of integers")
        return ell_m[:, 0] * (ell_m[:, 0] + 1) - self.ell_min ** 2 + ell_m[:, 1]

    @waveform_alterations
    def truncate(self, tol=1e-10):
        """Truncate the precision of this object's `data` in place

        This function sets bits in `self.data` to 0 when they will typically contribute
        less than `tol` times the norm of the Waveform at that instant in time.  Here,
        "typically" means that we divide `tol` by the square-root of the number of
        modes in the waveform, so that if each mode contributes a random amount of
        error at that level, we can expect a *total* error of roughly `tol` times the
        norm.

        """
        if tol != 0.0:
            tol_per_mode = tol / np.sqrt(self.n_modes)
            absolute_tolerance = np.linalg.norm(self.data, axis=1) * tol_per_mode
            power_of_2 = (2.0 ** np.floor(-np.log2(absolute_tolerance)))[:, np.newaxis]
            self.data *= power_of_2
            np.round(self.data, out=self.data)
            self.data /= power_of_2
        self._append_history(f"{self}.truncate(tol={tol})")

    def ladder_factor(self, operations, s, ell, eth_convention="NP"):
        """Compute the 'ladder factor' for applying a sequence of spin
        raising/lower operations on a SWSH of given s, ell.

        Parameters
        ----------
        operations: list or str composed of +1, -1, '+', '-', 'ð', or 'ð̅'
            The order of eth (+1, '+', or 'ð') and
            ethbar (-1, '-', or 'ð̅') operations to perform, applied from right to
            left. Example, operations='--+' will perform on WaveformModes data f the operation
            ethbar(ethbar(eth(f))).  Note that the order of operations is right-to-left.
        eth_convention: either 'NP' or 'GHP' [default: 'NP']
            Choose between Newman-Penrose or Geroch-Held-Penrose convention

        Returns
        -------
        float
        """

        op_dict = {"ð": +1, "ð̅": -1, "+": +1, "-": -1, +1: +1, -1: -1}
        conv_dict = {"NP": 1.0, "GHP": 0.5}

        # Normalize combining unicode characters
        if isinstance(operations, str):
            operations = operations.replace("ð̅", "-").replace("ð", "+")

        keys = op_dict.keys()
        key_strings = {key for key in keys if isinstance(key, str)}
        if not set(operations).issubset(keys):
            raise ValueError(
                "operations must be a string composed of "
                "{} or a list with "
                "elements coming from the set {}".format(key_strings, set(keys))
            )

        if eth_convention not in conv_dict:
            raise ValueError("eth_convention must be one of {}".format(set(conv_dict.keys())))

        convention_factor = conv_dict[eth_convention]

        ladder = 1.0
        sign_factor = 1.0

        for op in reversed(operations):
            sign = op_dict[op]
            sign_factor *= sign
            ladder *= (ell - s * sign) * (ell + s * sign + 1.0) if (ell >= abs(s)) else 0.0
            ladder *= convention_factor
            s += sign

        ladder = sign_factor * np.sqrt(ladder)
        return ladder

    def apply_eth(self, operations, eth_convention="NP"):
        """Apply spin raising/lowering operators to waveform mode data
        in a specified order.  This does not modify the original
        waveform object.

        Parameters
        ----------
        operations: list or str composed of +1, -1, '+', '-', 'ð', or 'ð̅'
            The order of eth (+1, '+', or 'ð') and
            ethbar (-1, '-', or 'ð̅') operations to perform, applied from right to
            left. Example, operations='--+' will perform on WaveformModes data f the operation
            ethbar(ethbar(eth(f))).  Note that the order of operations is right-to-left.
        eth_convention: either 'NP' or 'GHP' [default: 'NP']
            Choose between Newman-Penrose or Geroch-Held-Penrose convention

        Returns
        -------
        mode_data: array of complex
            Note that the returned data has the same shape as this object's `data` attribute, and
            the modes correspond to the same (ell, m) values.  In particular, if the spin weight
            changes, the output data will no longer satisfy the expectation that ell_min == abs(s).
        """

        s = self.spin_weight
        mode_data = self.data.copy()

        for ell in range(self.ell_min, self.ell_max + 1):
            ladder_factor = self.ladder_factor(operations, s, ell, eth_convention=eth_convention)
            lm_indices = [sf.LM_index(ell, m, self.ell_min) for m in range(-ell, ell + 1)]
            mode_data[:, lm_indices] *= ladder_factor

        return mode_data

    @property
    def eth(self):
        """Returns the spin-raised waveform mode data."""
        return self.apply_eth(operations="+")

    @property
    def ethbar(self):
        """Returns the spin-lowered waveform mode data."""
        return self.apply_eth(operations="-")

    def inner_product(self, b, t1=None, t2=None, allow_LM_differ=False, allow_times_differ=False):
        """Compute the all-angles inner product <self, b>.

        self and b should have the same spin-weight, set of (ell,m)
        modes, and times.  Differing sets of modes and times can
        optionally be turned on.

        Parameters
        ----------
        b : WaveformModes object
            The other set of modes to inner product with.

        t1 : float, optional [default: None]
        t2 : float, optional [default: None]
            Lower and higher bounds of time integration

        allow_LM_differ : bool, optional [default: False]
            If True and if the set of (ell,m) modes between self and b
            differ, then the inner product will be computed using the
            intersection of the set of modes.

        allow_times_differ: bool, optional [default: False]
            If True and if the set of times between self and b differ,
            then both WaveformModes will be interpolated to the
            intersection of the set of times.

        Returns
        -------
        inner_product : complex
            <self, b>
        """

        from quaternion.calculus import spline_definite_integral as sdi

        from .extrapolation import intersection

        if self.spin_weight != b.spin_weight:
            raise ValueError("Spin weights must match in inner_product")

        LM_clip = None
        if (self.ell_min != b.ell_min) or (self.ell_max != b.ell_max):
            if allow_LM_differ:
                LM_clip = slice(max(self.ell_min, b.ell_min), min(self.ell_max, b.ell_max) + 1)
                if LM_clip.start >= LM_clip.stop:
                    raise ValueError("Intersection of (ell,m) modes is " "empty.  Assuming this is not desired.")
            else:
                raise ValueError(
                    "ell_min and ell_max must match in inner_product " "(use allow_LM_differ=True to override)"
                )

        t_clip = None
        if not np.array_equal(self.t, b.t):
            if allow_times_differ:
                t_clip = intersection(self.t, b.t)
            else:
                raise ValueError(
                    "Time samples must match in inner_product " "(use allow_times_differ=True to override)"
                )

        ##########

        times = self.t
        A = self
        B = b

        if LM_clip is not None:
            A = A[:, LM_clip]
            B = B[:, LM_clip]

        if t_clip is not None:
            times = t_clip
            A = A.interpolate(t_clip)
            B = B.interpolate(t_clip)

        if t1 is None:
            t1 = times[0]

        if t2 is None:
            t2 = times[-1]

        integrand = np.sum(np.conj(A.data) * B.data, axis=1)

        return sdi(integrand, times, t1=t1, t2=t2)

    @waveform_alterations
    def convert_to_conjugate_pairs(self):
        """Convert modes to conjugate-pair format in place

        This function alters this object's modes to store the sum and difference of pairs with
        opposite `m` values.  If we denote the modes `f[l, m]`, then we define

            s[l, m] = (f[l, m] + f̄[l, -m]) / √2
            d[l, m] = (f[l, m] - f̄[l, -m]) / √2

        For m<0 we replace the mode data with `d[l, -m]`, for m=0 we do nothing, and for m>0 we
        replace the mode data with `s[l, m]`.  That is, the mode data on output look like this:

            [..., d[2, 2], d[2, 1], f[2, 0], s[2, 1], s[2, 2], d[3, 3], d[3, 2], ...]

        The factor of √2 is chosen so that the sum of the magnitudes squared at each time for this
        data is the same as it is for the original data.

        """
        for ell in range(self.ell_min, self.ell_max + 1):
            for m in range(1, ell + 1):
                i_plus = self.index(ell, m)
                i_minus = self.index(ell, -m)
                mode_plus = self.data[..., i_plus].copy()
                mode_minus = self.data[..., i_minus].copy()
                self.data[..., i_plus] = (mode_plus + np.conjugate(mode_minus)) / np.sqrt(2)
                self.data[..., i_minus] = (mode_plus - np.conjugate(mode_minus)) / np.sqrt(2)
        self._append_history(f"{self}.convert_to_conjugate_pairs()")

    @waveform_alterations
    def convert_from_conjugate_pairs(self):
        """Convert modes from conjugate-pair format in place

        This function reverses the effects of `convert_to_conjugate_pairs`.  See that function's
        docstring for details.

        """
        for ell in range(self.ell_min, self.ell_max + 1):
            for m in range(1, ell + 1):
                i_plus = self.index(ell, m)
                i_minus = self.index(ell, -m)
                mode_plus = self.data[..., i_plus].copy()
                mode_minus = self.data[..., i_minus].copy()
                self.data[..., i_plus] = (mode_plus + mode_minus) / np.sqrt(2)
                self.data[..., i_minus] = np.conjugate(mode_plus - mode_minus) / np.sqrt(2)
        self._append_history(f"{self}.convert_from_conjugate_pairs()")

    def transform(self, **kwargs):
        """Transform modes by some BMS transformation

        This simply applies the `WaveformGrid.from_modes` function, followed by the
        `WaveformGrid.to_modes` function.  See their respective docstrings for more
        details.

        However, note that the `ell_max` parameter used in the second function call
        defaults here to the `ell_max` value in the input waveform.  This is slightly
        different from the usual default, because `WaveformGrid.from_modes` usually
        increases the effective ell value by 1.

        """
        from . import WaveformGrid
        return WaveformGrid.transform(self, **kwargs)

    # Involutions
    @property
    @waveform_alterations
    def x_parity_conjugate(self):
        """Reflect modes across y-z plane (along x axis)

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError(f"Cannot compute parity type for {self.data_type_string}.")
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.empty_like(self.data)
        W.ells = self.ells
        W.frame = np.x_parity_conjugate(self.frame)
        for ell in range(W.ell_min, W.ell_max + 1):
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if (m % 2) == 0]
            W.data[:, lm_indices] = np.conjugate(self.data[:, lm_indices])
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if (m % 2) != 0]
            W.data[:, lm_indices] = -np.conjugate(self.data[:, lm_indices])
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.x_parity_conjugate")
        return W

    @property
    @waveform_alterations
    def x_parity_symmetric_part(self):
        """Return component of waveform invariant under `x_parity_conjugate`"""
        W = self.x_parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.x_parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.x_parity_symmetric_part")
        return W

    @property
    @waveform_alterations
    def x_parity_antisymmetric_part(self):
        """Return component of waveform that changes sign under `x_parity_conjugate`"""
        W = self.x_parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.x_parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.x_parity_antisymmetric_part")
        return W

    @property
    def x_parity_violation_squared(self):
        """(Squared) norm of x-parity-antisymmetric component of waveform"""
        return self.x_parity_antisymmetric_part.norm()

    @property
    def x_parity_violation_normalized(self):
        """Norm of x-parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.x_parity_antisymmetric_part.norm() / self.norm())

    @property
    @waveform_alterations
    def y_parity_conjugate(self):
        """Reflect modes across x-z plane (along y axis)

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError(f"Cannot compute parity type for {self.data_type_string}.")
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.conjugate(self.data)
        W.ells = self.ells
        W.frame = np.y_parity_conjugate(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.y_parity_conjugate")
        return W

    @property
    @waveform_alterations
    def y_parity_symmetric_part(self):
        """Component of waveform invariant under `y_parity_conjugate`"""
        W = self.y_parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.y_parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.y_parity_symmetric_part")
        return W

    @property
    @waveform_alterations
    def y_parity_antisymmetric_part(self):
        """Component of waveform that changes sign under `y_parity_conjugate`"""
        W = self.y_parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.y_parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.y_parity_antisymmetric_part")
        return W

    @property
    def y_parity_violation_squared(self):
        """(Squared) norm of y-parity-antisymmetric component of waveform"""
        return self.y_parity_antisymmetric_part.norm()

    @property
    def y_parity_violation_normalized(self):
        """Norm of y-parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.y_parity_antisymmetric_part.norm() / self.norm())

    @property
    @waveform_alterations
    def z_parity_conjugate(self):
        """Reflect modes across x-y plane (along z axis)

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError(f"Cannot compute parity type for {self.data_type_string}.")
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.empty_like(self.data)
        W.ells = self.ells
        W.frame = np.z_parity_conjugate(self.frame)
        s = self.spin_weight
        for ell in range(W.ell_min, W.ell_max + 1):
            if ((ell + s) % 2) == 0:
                lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1)]
                W.data[:, lm_indices] = np.conjugate(self.data[:, list(reversed(lm_indices))])
            else:
                lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1)]
                W.data[:, lm_indices] = -np.conjugate(self.data[:, list(reversed(lm_indices))])
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.z_parity_conjugate")
        return W

    @property
    @waveform_alterations
    def z_parity_symmetric_part(self):
        """Component of waveform invariant under `z_parity_conjugate`"""
        W = self.z_parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.z_parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.z_parity_symmetric_part")
        return W

    @property
    @waveform_alterations
    def z_parity_antisymmetric_part(self):
        """Component of waveform that changes sign under `z_parity_conjugate`"""
        W = self.z_parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.z_parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.z_parity_antisymmetric_part")
        return W

    @property
    def z_parity_violation_squared(self):
        """(Squared) norm of z-parity-antisymmetric component of waveform"""
        return self.z_parity_antisymmetric_part.norm()

    @property
    def z_parity_violation_normalized(self):
        """Norm of z-parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.z_parity_antisymmetric_part.norm() / self.norm())

    @property
    @waveform_alterations
    def parity_conjugate(self):
        """Reflect modes along all axes

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError(f"Cannot compute parity type for {self.data_type_string}.")
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.empty_like(self.data)
        W.ells = self.ells
        W.frame = np.parity_conjugate(self.frame)
        s = self.spin_weight
        for ell in range(W.ell_min, W.ell_max + 1):
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if ((ell + s + m) % 2) == 0]
            W.data[:, lm_indices] = np.conjugate(self.data[:, list(reversed(lm_indices))])
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if ((ell + s + m) % 2) != 0]
            W.data[:, lm_indices] = -np.conjugate(self.data[:, list(reversed(lm_indices))])
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.parity_conjugate")
        return W

    @property
    @waveform_alterations
    def parity_symmetric_part(self):
        """Component of waveform invariant under `parity_conjugate`"""
        W = self.parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.parity_symmetric_part")
        return W

    @property
    @waveform_alterations
    def parity_antisymmetric_part(self):
        """Component of waveform that changes sign under `parity_conjugate`"""
        W = self.parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.parity_antisymmetric_part")
        return W

    @property
    def parity_violation_squared(self):
        """(Squared) norm of parity-antisymmetric component of waveform"""
        return self.parity_antisymmetric_part.norm()

    @property
    def parity_violation_normalized(self):
        """Norm of parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.parity_antisymmetric_part.norm() / self.norm())

    @waveform_alterations
    def copy_without_data(self):
        W = super().copy_without_data()
        W.ells = 0, -1
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}.copy_without_data()")
        return W

    @waveform_alterations
    def __getitem__(self, key):
        """Extract subsets of the data efficiently

        Note that if a second index is given to this function, it corresponds to modes, rather than data indices. See
        the docstring of the WaveformModes class for examples.

        """
        # Remove trivial tuple structure first
        if isinstance(key, tuple) and len(key) == 1:
            key = key[0]

        # Now figure out which type of return is desired
        if isinstance(key, tuple) and len(key) == 2:
            # Return a subset of the data from a subset of times
            if isinstance(key[1], int):
                if key[1] < self.ell_min or key[1] > self.ell_max:
                    raise ValueError(
                        "Requested ell value {} lies outside ".format(key[1])
                        + f"WaveformModes object's ell range ({self.ell_min},{self.ell_max})."
                    )
                new_ell_min = key[1]
                new_ell_max = key[1]
                new_slice = slice(
                    new_ell_min ** 2 - self.ell_min ** 2, new_ell_max * (new_ell_max + 2) + 1 - self.ell_min ** 2
                )
            elif isinstance(key[1], slice):
                if key[1].step and key[1].step != 1:
                    raise ValueError(
                        "Can only slice WaveformModes over contiguous ell values (step={})".format(key[1].step)
                    )
                if not key[1].start and key[1].stop == 0:
                    new_ell_min = 0
                    new_ell_max = -1
                    new_slice = slice(0)
                else:
                    if not key[1].start:
                        new_ell_min = self.ell_min
                    else:
                        new_ell_min = key[1].start
                    if not key[1].stop:
                        new_ell_max = self.ell_max
                    else:
                        new_ell_max = key[1].stop - 1
                    if new_ell_min < self.ell_min or new_ell_max > self.ell_max:
                        raise ValueError(
                            f"Requested ell range [{new_ell_min},{new_ell_max}] lies outside "
                            + f"WaveformBase's ell range [{self.ell_min},{self.ell_max}]."
                        )
                    new_slice = slice(
                        new_ell_min ** 2 - self.ell_min ** 2, new_ell_max * (new_ell_max + 2) + 1 - self.ell_min ** 2
                    )
            else:
                raise ValueError("Don't know what to do with slice of type `{}`".format(type(key[1])))
            W = super().__getitem__((key[0], new_slice))
            W.ells = new_ell_min, new_ell_max
        elif isinstance(key, slice) or isinstance(key, int):
            # Return complete data from a subset of times (key is slice), or
            # return complete data from a single instant in time (key is int)
            W = super().__getitem__(key)
            W.ells = self.ells
        else:
            raise ValueError("Could not understand input `{}` (of type `{}`) ".format(key, type(key)))

        W.history.pop()  # remove WaveformBase history append, to be replaced next
        W.__history_depth__ -= 1
        W._append_history(f"{W} = {self}[{key}]")

        return W

    def __repr__(self):
        # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        rep = super().__repr__()
        rep += f"\n# ell_min={self.ell_min}, ell_max={self.ell_max}"
        return rep

    # N.B.: There are additional methods attached to `WaveformModes` in the `waveform_base` file.  These functions
    # cannot be added here, because they depend on `WaveformGrid` objects, which depend on `WaveformModes` objects.



# Copyright (c) 2021, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

import warnings
import os
import re
import ast
import numpy as np
import h5py
import quaternion
import spherical_functions as sf
from ... import (
    jit,
    WaveformModes,
    FrameNames,
    DataType,
    DataNames,
    UnknownDataType,
    h,
    hdot,
    psi4,
    psi3,
    psi2,
    psi1,
    psi0,
    Inertial,
)
import sxs
from sxs.metadata import Metadata
from . import corotating_paired_xor, rotating_paired_xor_multishuffle_bzip2
from ...asymptotic_bondi_data import AsymptoticBondiData

rpxmb = rotating_paired_xor_multishuffle_bzip2


def translate_data_types_GWFrames_to_waveforms(d):
    if d < 8:
        return {0: UnknownDataType, 1: h, 2: hdot, 3: psi4, 4: psi3, 5: psi2, 6: psi1, 7: psi0}[d]
    else:
        return DataType[d - 8]


def translate_data_types_waveforms_to_GWFrames(d):
    if d in [UnknownDataType, h, hdot, psi4, psi3, psi2, psi1, psi0]:
        return {UnknownDataType: 0, h: 1, hdot: 2, psi4: 3, psi3: 4, psi2: 5, psi1: 6, psi0: 7}[d]
    else:
        return d + 8


@jit
def index_is_monotonic(y):
    length = y.size
    monotonic = np.ones_like(y, dtype=np.bool_)
    direction = y[-1] - y[0]
    if direction > 0.0:
        max_value = y[0]
        for i in range(1, length):
            if y[i] <= max_value:
                monotonic[i] = False
            else:
                max_value = y[i]
    else:
        min_value = y[0]
        for i in range(1, length):
            if y[i] >= min_value:
                monotonic[i] = False
            else:
                min_value = y[i]
    return monotonic


def monotonic_indices(y):
    indices = np.arange(y.size)
    return indices[index_is_monotonic(y)]


def monotonize(y):
    return y[monotonic_indices(y)]


def read_from_h5(file_name, **kwargs):
    """Read data from an H5 file in SXS format

    Note that SXS format is essentially compatible with NRAR format.  The existence of this function is not to be
    taken as the author's endorsement of either SXS or NRAR format.

    Parameters
    ----------
    file_name : str
        Path to H5 file containing the data, optionally including the path within the file itself to the directory
        containing the data.  For example, the standard SXS data with N=2 might be obtained with the file name
        `'rhOverM_Asymptotic_GeometricUnits.h5/Extrapolated_N2.dir'`.

    Keyword parameters
    ------------------
    frameType : int, optional
    dataType : int, optional
    r_is_scaled_out : bool, optional
    m_is_scaled_out : bool, optional
        These four parameters are documented in the docstring of the WaveformBase object.  Note that if any of these
        is present in the H5 file (which is not common) that value will override this argument.  If neither the file
        nor these parameters are present, defaults will be applied, assuming that the frame is inertial, R and M are
        both scaled out, and the data type (hdot, h, psi4, psi4, psi2, psi1, or psi0) can be gleaned from `file_name`.

    See also
    --------
    scri.rpxmb.save : Save a waveform in RPXMB format
    scri.rpxmb.load : Load a file in RPXMB format

    """

    import os.path
    import re
    import h5py
    import quaternion

    # This unfortunate concoction is needed to determine the (ell,m) values of the various mode data sets
    pattern_Ylm = re.compile(r"""Y_l(?P<L>[0-9]+)_m(?P<M>[-+0-9]+)\.dat""")

    # Initialize an empty object to be filled with goodies
    w = WaveformModes(constructor_statement=f'scri.SpEC.read_from_h5("{file_name}", **{kwargs})')

    # Get an h5py handle to the desired part of the h5 file
    try:
        file_name, root_group = file_name.rsplit(".h5", 1)
        file_name += ".h5"
    except ValueError:
        root_group = ""  # FileName is just a file, not a group in a file
    try:
        f_h5 = h5py.File(file_name, "r")
    except OSError:
        print(f"\n`read_from_h5` could not open the file '{file_name}'\n\n")
        raise
    if root_group:
        f = f_h5[root_group]
    else:
        f = f_h5

    # If it exists, add the metadata file to `w` as an object.  So, for example, the initial spin on
    # object 1 can be accessed as `w.metadata.initial_spin1`.  See the documentation of
    # `sxs.metadata.Metadata` for more details.  And in IPython, tab completion works on the
    # `w.metadata` object.
    try:
        w.metadata = Metadata.from_file(
            os.path.join(os.path.dirname(file_name), "metadata"), ignore_invalid_lines=True, cache_json=False
        )
    except:
        pass  # Probably couldn't find the metadata.json/metadata.txt file

    try:  # Make sure the h5py.File gets closed, even in the event of an exception
        # Add the old history to the new history, if found
        try:
            try:
                old_history = f["History.txt"][()].decode()
            except AttributeError:
                old_history = f["History.txt"][()]
            w._append_history("", 0)
            w._append_history("<previous_history>", 2)
            w._append_history(old_history, 1)
            w._append_history("</previous_history>", 2)
        except KeyError:
            old_history = False

        # Get the frame data, converting to quaternion objects
        try:
            w.frame = quaternion.as_quat_array(f["Frame"])
        except KeyError:
            pass  # There was no frame data

        # Get the descriptive items
        try:
            w.frameType = int(f.attrs["FrameType"])
        except KeyError:
            if "frameType" in kwargs:
                w.frameType = int(kwargs.pop("frameType"))
            else:
                warning = (
                    f"\n`frameType` was not found in '{file_name}' or the keyword arguments.\n"
                    + "Using default value `{}`.  You may want to set it manually.\n\n".format(FrameNames[1])
                )
                warnings.warn(warning)
                w.frameType = 1
        try:
            w.dataType = translate_data_types_GWFrames_to_waveforms(int(f.attrs["DataType"]))
        except KeyError:
            if "dataType" in kwargs:
                w.dataType = int(kwargs.pop("dataType"))
            else:
                found = False
                for type_int, type_name in zip(reversed(DataType), reversed(DataNames)):
                    if type_name.lower() in file_name.lower():
                        found = True
                        w.dataType = type_int
                        warning = (
                            f"\n`dataType` was not found in '{file_name}' or the keyword arguments.\n"
                            + f"Using default value `{type_name}`.  You may want to set it manually."
                        )
                        warnings.warn(warning)
                        break
                if not found:
                    warning = (
                        f"\n`dataType` was not found in '{file_name}' or the keyword arguments.\n"
                        + "You may want to set it manually."
                    )
                    warnings.warn(warning)
        try:
            w.r_is_scaled_out = bool(f.attrs["RIsScaledOut"])
        except KeyError:
            if "r_is_scaled_out" in kwargs:
                w.r_is_scaled_out = bool(kwargs.pop("r_is_scaled_out"))
            else:
                warning = (
                    f"\n`r_is_scaled_out` was not found in '{file_name}' or the keyword arguments.\n"
                    + "Using default value `True`.  You may want to set it manually.\n\n"
                )
                warnings.warn(warning)
                w.r_is_scaled_out = True
        try:
            w.m_is_scaled_out = bool(f.attrs["MIsScaledOut"])
        except KeyError:
            if "m_is_scaled_out" in kwargs:
                w.m_is_scaled_out = bool(kwargs.pop("m_is_scaled_out"))
            else:
                warning = (
                    f"\n`m_is_scaled_out` was not found in '{file_name}' or the keyword arguments.\n"
                    + "Using default value `True`.  You may want to set it manually.\n\n"
                )
                warnings.warn(warning)
                w.m_is_scaled_out = True

        # Get the names of all the data sets in the h5 file, and check for matches
        YLMdata = [data_set for data_set in list(f) for m in [pattern_Ylm.search(data_set)] if m]
        if len(YLMdata) == 0:
            raise ValueError(
                f"Couldn't understand data set names in '{file_name}'.\n"
                + "Maybe you need to add the directory within the h5 file.\n"
                + f"E.g.: '{file_name}/Extrapolated_N2.dir'."
            )

        # Sort the data set names by increasing ell, then increasing m
        YLMdata = sorted(
            YLMdata,
            key=lambda data_set: [
                int(pattern_Ylm.search(data_set).group("L")),
                int(pattern_Ylm.search(data_set).group("M")),
            ],
        )
        LM = np.array(
            sorted(
                [
                    [int(m.group("L")), int(m.group("M"))]
                    for data_set in YLMdata
                    for m in [pattern_Ylm.search(data_set)]
                    if m
                ]
            )
        )
        ell_min, ell_max = min(LM[:, 0]), max(LM[:, 0])
        if not np.array_equal(LM, sf.LM_range(ell_min, ell_max)):
            raise ValueError(f"Input [ell,m] modes are not complete.  Found modes:\n{LM}\n")
        n_modes = len(LM)

        # Get the time data (assuming all are equal)
        T = f[YLMdata[0]][:, 0]
        monotonic = index_is_monotonic(T)
        w.t = T[monotonic]
        n_times = len(w.t)

        # Loop through, setting data in each mode
        w.data = np.empty((n_times, n_modes), dtype=complex)
        for m, DataSet in enumerate(YLMdata):
            if f[DataSet].shape[0] != n_times:
                raise ValueError(
                    f"The number of time steps in this dataset should be {n_times}; "
                    + "it is {} in '{}'.".format(f[DataSet].shape[0], DataSet)
                )
            w.data[:, m] = f[DataSet][:, 1:3].view(dtype=complex)[monotonic, 0]

        # Now that the data is set, we can set these
        w.ells = ell_min, ell_max

        # If possible, retrieve the CoM-correction parameters
        try:
            if hasattr(f, "attrs") and "space_translation" in f.attrs:
                w.space_translation = np.array(list(f.attrs["space_translation"]))
            elif old_history:
                pattern = r"'{}': array\((.*?)\)".format("space_translation")
                matches = re.search(pattern, old_history)
                if matches:
                    w.space_translation = np.array(ast.literal_eval(matches.group(1)))
        except:
            pass
        try:
            if hasattr(f, "attrs") and "boost_velocity" in f.attrs:
                w.boost_velocity = np.array(list(f.attrs["boost_velocity"]))
            elif old_history:
                pattern = r"'{}': array\((.*?)\)".format("boost_velocity")
                matches = re.search(pattern, old_history)
                if matches:
                    w.boost_velocity = np.array(ast.literal_eval(matches.group(1)))
        except:
            pass

        # If possible, retrieve the CoM-correction parameters
        try:
            if "VersionHist.ver" in f_h5:
                w.version_hist = [
                    [git_hash.decode("ascii"), message.decode("ascii")]
                    for git_hash, message in f_h5["VersionHist.ver"][()].tolist()
                ]
        except:
            pass

        # Check up on the validity of the waveform
        if not w.ensure_validity(alter=True, assertions=False):
            raise ValueError("The data resulting from this input is invalid")

    except KeyError:
        print("\nThis H5 file appears to have not stored all the required information.\n\n")
        raise  # Re-raise the exception after adding our information

    finally:  # Use `finally` to make sure this happens:
        f_h5.close()

    if kwargs:
        import pprint

        warnings.warn("\nUnused kwargs passed to this function:\n{}".format(pprint.pformat(kwargs, width=1)))

    return w


def write_to_h5(w, file_name, file_write_mode="w", attributes={}, use_NRAR_format=True):
    """
    Output the Waveform to an HDF5 file. Default behavior uses the NRAR format.

    Note that the file_name is prepended with some descriptive information involving the data type and the frame type,
    such as 'rhOverM_Corotating_' or 'rMpsi4_Aligned_'.

    See also
    --------
    scri.rpxmb.save : Save a waveform in RPXMB format
    scri.rpxmb.load : Load a file in RPXMB format

    """

    import os.path
    import h5py
    import warnings

    group = None
    if ".h5" in file_name and not file_name.endswith(".h5"):
        file_name, group = file_name.split(".h5")
        file_name += ".h5"
    # Add descriptive prefix to file_name
    base_name = w.descriptor_string + "_" + os.path.basename(file_name)
    if not os.path.dirname(file_name):
        file_name = base_name
    else:
        file_name = os.path.join(os.path.dirname(file_name), base_name)
    # Open the file for output
    try:
        f = h5py.File(file_name, file_write_mode)
    except OSError:  # If that did not work...
        print(f"write_to_h5 was unable to open the file '{file_name}'.\n\n")
        raise  # re-raise the exception after the informative message above
    try:
        # If we are writing to a group within the file, create it
        if group:
            g = f.create_group(group)
        else:
            g = f
        # Now write all the data to various groups in the file
        g.attrs["OutputFormatVersion"] = "scri.SpEC"
        g.create_dataset("History.txt", data="\n".join(w.history) + f"\n\nwrite_to_h5({w}, {file_name})\n")
        g.attrs["FrameType"] = w.frameType
        g.attrs["DataType"] = translate_data_types_waveforms_to_GWFrames(w.dataType)
        g.attrs["RIsScaledOut"] = int(w.r_is_scaled_out)
        g.attrs["MIsScaledOut"] = int(w.m_is_scaled_out)
        if len(w.version_hist) > 0:
            try:
                version_hist = [[e.encode("ascii", "ignore") for e in hm] for hm in w.version_hist]
            except AttributeError:
                version_hist = w.version_hist
            g.create_dataset("VersionHist.ver", data=version_hist)
        for attr in attributes:
            try:
                g.attrs[attr] = attributes[attr]
            except:
                warning = f"scri.SpEC.write_to_h5 unable to output attribute {attr}={attributes[attr]}"
                warnings.warn(warning)
        if use_NRAR_format:
            for i_m in range(w.n_modes):
                ell, m = w.LM[i_m]
                Data_m = g.create_dataset(
                    f"Y_l{ell}_m{m}.dat",
                    data=[[t, d.real, d.imag] for t, d in zip(w.t, w.data[:, i_m])],
                    compression="gzip",
                    shuffle=True,
                )
                Data_m.attrs["ell"] = ell
                Data_m.attrs["m"] = m
        else:
            g.create_dataset("Time", data=w.t.tolist(), compression="gzip", shuffle=True)
            if len(w.frame) > 0:
                g.create_dataset("Frame", data=[[r.w, r.x, r.y, r.z] for r in w.frame])
            else:
                g.create_dataset("Frame", shape=())
            Data = g.create_group("Data")
            for i_m in range(w.n_modes):
                ell, m = w.LM[i_m]
                Data_m = Data.create_dataset(
                    "l{}_m{:+}".format(int(ell), int(m)), data=w.data[:, i_m], compression="gzip", shuffle=True
                )
                Data_m.attrs["ell"] = ell
                Data_m.attrs["m"] = m
    finally:  # Use `finally` to make sure this happens:
        f.close()


def make_variable_dimensionless(WM, ch_mass=1.0):
    """
    Make variable dimensionless by scaling out Christodoulou mass.

    Parameters:
    -----------
    WM : scri.WaveformModes
        Waveform data.
    ch_mass : float, optional
        Total Christodoulou mass of the system.
        [Default: 1.0].

    """

    if WM.m_is_scaled_out:
        print("Data is already dimensionless!")
        return

    if WM.dataType in [psi4, psi3, psi2, psi1, psi0]:
        unit_scale_factor = (ch_mass) ** (WM.dataType - 4)
    elif WM.dataType == h:
        unit_scale_factor = 1 / ch_mass
    elif WM.dataType == hdot:
        unit_scale_factor = 1.0
    else:
        raise ValueError("DataType not determined.")

    WM.t = WM.t / ch_mass
    WM.data = WM.data * unit_scale_factor
    WM.m_is_scaled_out = True


def create_abd_from_h5(
    file_format,
    convention="SpEC",
    radius=None,
    ch_mass=None,
    t_interpolate=None,
    t_0_superrest=None,
    padding_time=None,
    **kwargs,
):
    """Returns an AsymptoticBondiData object with waveform data loaded from specified H5 files.

    The AsymptoticBondiData class internally uses the Moreschi-Boyle conventions, see the following reference:
      O. Moreschi, On angular momentum at future null infinity, DOI:10.1088/0264-9381/3/4/006
    If necessary, the waveform data will be converted to the Moreschi-Boyle conventions when loaded.

    Parameters
    ----------
    file_format : 'SXS', 'SpECTRECCE_v1', 'RPDMB', or 'RPXMB'
        The H5 files may be in the one of the following file formats:
          * 'SXS'  - Dimensionless extrapolated waveform files found in the SXS Catalog, also known as NRAR format.
          * 'SpECTRECCE_v1' - Asymptotic waveforms output by SpECTRE CCE v1.
          * 'RPDMB' - Dimensionless waveforms compressed using the rotating_paired_diff_multishuffle_bzip2 format.
          * 'RPXMB' - Dimensionless waveforms compressed using the rotating_paired_xor_multishuffle_bzip2 format.
    convention : 'SpEC' or 'Moreschi-Boyle'
        The data conventions of the waveform data that will be loaded. This defaults to 'SpEC' since this will be
        most often used with 'SpEC' convention waveforms. The output convention is Moreschi-Boyle.
    radius : str, optional
        Worldtube radius used when running CCE; only needed for versions of SpECTRE before PR #5985.
        The time array of the worldtube is translated by this radius.
    ch_mass : float, optional
        Total Christodoulou mass of the system.
    t_interpolate : float array, optional
        Time array to interpolate to, e.g., the time array of the worldtube.
    t_0_superrest : float, optional
        When to map to the BMS superrest frame.
        Typically a few hundred M after the junk radiation is sufficient.
    padding_time : float, optional
        Time window length around t_0_superrest to use when mapping to the superrest frame.
        Typically a few hundred M or a few orbits is sufficient.

    Keyword Parameters
    ------------------
    Psi4 : str, optional
    Psi3 : str, optional
    Psi2 : str, optional
    Psi1 : str, optional
    Psi0 : str, optional
    h    : str, optional
        Path to H5 file containing the data. At least ONE the above waveform quantities is required.


    Returns
    -------
    AsymptoticBondiData

    """

    # Use case insensitive parameters
    file_format = file_format.lower()
    convention = convention.lower()

    # Load waveform data from H5 files into WaveformModes objects
    WMs = {}
    filenames = {}
    if file_format == "spectrecce_v1":
        try:
            file_name = kwargs.pop("file_name")
        except:
            raise ValueError('Need to specify "file_name" option!')

        with h5py.File(file_name, "r") as f:
            for x in f.keys():
                if "Spectre" in x:
                    cce_key = x
                    radius = cce_key.split("R")[1][:4]
                    data_label_suffix = ""
                    break
                else:
                    cce_key = "Cce"
                    data_label_suffix = ".dat"

            cce = f[cce_key]
            time = cce[f"Strain{data_label_suffix}"][:, 0]
            indices = monotonic_indices(time)
            time = time[indices]
            ell_max = int(np.sqrt((cce[f"Strain{data_label_suffix}"].shape[1] - 1) / 2) - 1)
            for data_label in ["Psi4", "Psi3", "Psi2", "Psi1", "Psi0", "Strain"]:
                if data_label != "Strain":
                    dataType = DataNames.index(data_label)
                else:
                    dataType = DataNames.index("h")
                WMs[data_label] = WaveformModes(
                    t=time,
                    data=cce[f"{data_label}{data_label_suffix}"][indices, 1:].view(np.complex128),
                    ell_min=0,
                    ell_max=ell_max,
                    frameType=Inertial,
                    dataType=dataType,
                    constructor_statement=f"create_abd_from_h5({file_format}, {convention=}, {file_name=})",
                    r_is_scaled_out=True,
                    m_is_scaled_out=False,
                )
    else:
        for data_label in ["Psi4", "Psi3", "Psi2", "Psi1", "Psi0", "h"]:
            if data_label in kwargs:
                filenames[data_label] = kwargs.pop(data_label)
                if file_format == "sxs":
                    WMs[data_label] = read_from_h5(filenames[data_label])
                elif file_format == "cce":
                    WMs[data_label] = read_from_h5(
                        filenames[data_label],
                        dataType=DataNames.index(data_label),
                        frameType=Inertial,
                        r_is_scaled_out=True,
                        m_is_scaled_out=False,
                    )
                elif file_format == "rpxmb" or file_format == "rpxm":
                    WMs[data_label] = rotating_paired_xor_multishuffle_bzip2.load(filenames[data_label])[0]
                    WMs[data_label].to_inertial_frame()
                elif file_format == "rpdmb" or file_format == "rpdm":
                    WMs[data_label] = WaveformModes.from_sxs(sxs.rpdmb.load(filenames[data_label]))
                else:
                    raise ValueError(
                        f"File format '{file_format}' not recognized. "
                        "Must be 'SXS', 'CCE', 'SpECTRECCE_v1', 'RPDMB', or 'RPXMB'."
                    )

    if kwargs:
        import pprint

        warnings.warn("\nUnused kwargs passed to this function:\n{}".format(pprint.pformat(kwargs, width=1)))

    # Sanity check
    if not WMs:
        raise ValueError("No filenames have been provided. The data of at least one waveform quantity is required.")

    WM_ref = WMs[list(WMs.keys())[0]]
    for i in WMs:
        if not (WM_ref.t == WMs[i].t).all():
            raise ValueError(
                f"All waveforms must share the same set of times. The data "
                f"for {list(WMs.keys())[i].data_type_string} has a different set of times."
            )

    for i in WMs:
        # Make waveforms dimensionless (if they already are, does nothing)
        if ch_mass is not None:
            make_variable_dimensionless(WMs[i], ch_mass)

        # Adjust time by worldtube radius
        if file_format == "spectrecce_v1":
            WMs[i].t -= float(radius)

    # Create an instance of AsymptoticBondiData
    abd = AsymptoticBondiData(
        time=WM_ref.t,
        ell_max=WM_ref.ell_max,
        multiplication_truncator=max,
    )

    # Define factors to convert between input waveform convention and Moreschi-Boyle convention
    conversion_factor = {
        # "input convention" : [Ψ₀, Ψ₁, Ψ₂, Ψ₃, Ψ₄, h]
        "moreschi-boyle": [1, 1, 1, 1, 1, 1],
        "spec": [2, -np.sqrt(2), 1, -1 / np.sqrt(2), 0.5, 0.5],
    }

    # Load the WaveformModes data into the ABD object and convert to the Moreschi-Boyle convention.
    if "Psi4" in WMs:
        abd.psi4[:, sf.LM_index(WMs["Psi4"].ell_min, -WMs["Psi4"].ell_min, 0) :] = (
            conversion_factor[convention][4] * WMs["Psi4"].data
        )
    if "Psi3" in WMs:
        abd.psi3[:, sf.LM_index(WMs["Psi3"].ell_min, -WMs["Psi3"].ell_min, 0) :] = (
            conversion_factor[convention][3] * WMs["Psi3"].data
        )
    if "Psi2" in WMs:
        abd.psi2[:, sf.LM_index(WMs["Psi2"].ell_min, -WMs["Psi2"].ell_min, 0) :] = (
            conversion_factor[convention][2] * WMs["Psi2"].data
        )
    if "Psi1" in WMs:
        abd.psi1[:, sf.LM_index(WMs["Psi1"].ell_min, -WMs["Psi1"].ell_min, 0) :] = (
            conversion_factor[convention][1] * WMs["Psi1"].data
        )
    if "Psi0" in WMs:
        abd.psi0[:, sf.LM_index(WMs["Psi0"].ell_min, -WMs["Psi0"].ell_min, 0) :] = (
            conversion_factor[convention][0] * WMs["Psi0"].data
        )
    # ABD uses the Newman-Penrose scalar sigma instead of the strain h, so we have to take
    # the complex conjugate.
    if "h" in WMs:
        abd.sigma[:, sf.LM_index(WMs["h"].ell_min, -WMs["h"].ell_min, 0) :] = (
            conversion_factor[convention][5] * WMs["h"].data
        )
        abd.sigma = abd.sigma.bar
    elif "Strain" in WMs:
        abd.sigma[:, sf.LM_index(WMs["Strain"].ell_min, -WMs["Strain"].ell_min, 0) :] = (
            conversion_factor[convention][5] * WMs["Strain"].data
        )
        abd.sigma = abd.sigma.bar

    # Interpolate to finer time array, if specified
    if t_interpolate is not None:
        idx1 = np.argmin(abs(t_interpolate - abd.t[0])) + 1
        idx2 = np.argmin(abs(t_interpolate - abd.t[-1])) + 1 - 1
        abd = abd.interpolate(t_interpolate[idx1:idx2])

    # Map to superrest frame at some time over some window, if specified
    if t_0_superrest is not None and padding_time is not None:
        abd, BMS, _ = abd.map_to_superrest_frame(t_0=t_0_superrest, padding_time=padding_time)

    return abd



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

import numpy as np
from matplotlib import scale as mscale
from matplotlib import transforms as mtransforms
from matplotlib.ticker import MaxNLocator


class WaveformTimeScale(mscale.ScaleBase):
    name = "merger_zoom"

    def __init__(self, axis, **kwargs):
        mscale.ScaleBase.__init__(self, axis)
        self.t_initial = float(kwargs.pop("t_initial", 0.0))
        self.t_merger = float(kwargs.pop("t_merger"))
        self.t_ringdown = float(kwargs.pop("t_ringdown"))
        self.t_final = float(kwargs.pop("t_final"))
        self.merger_stretch = float(
            kwargs.pop("merger_stretch", 0.75 * (self.t_final - self.t_initial) / (self.t_ringdown - self.t_merger))
        )

    def get_transform(self):
        return self.WaveformTimeTransform(
            self.merger_stretch, self.t_initial, self.t_merger, self.t_ringdown, self.t_final
        )

    def set_default_locators_and_formatters(self, axis):
        class StretchingLocator(MaxNLocator):
            def __init__(self, t_initial, t_merger, t_ringdown, t_final):
                MaxNLocator.__init__(self, nbins=9, steps=[1, 2, 5, 10])
                self.t_initial = t_initial
                self.t_merger = t_merger
                self.t_ringdown = t_ringdown
                self.t_final = t_final

            def tick_values(self, vmin, vmax):
                self.set_params(nbins=6, prune=None)
                ticks = list(super().tick_values(max(vmin, self.t_initial), min(vmax, self.t_merger)))
                self.set_params(nbins=11 - len(ticks), prune="both")
                ticks += list(super().tick_values(max(vmin, self.t_merger), min(vmax, self.t_ringdown)))
                self.set_params(nbins=9, prune=None)
                return np.array(ticks)

        axis.set_major_locator(StretchingLocator(self.t_initial, self.t_merger, self.t_ringdown, self.t_final))

    def limit_range_for_scale(self, vmin, vmax, minpos):
        return max(vmin, self.t_initial), min(vmax, self.t_final)

    class WaveformTimeTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def __init__(self, merger_stretch, t_initial, t_merger, t_ringdown, t_final):
            mtransforms.Transform.__init__(self)
            self.merger_stretch, self.t_initial, self.t_merger, self.t_ringdown, self.t_final = (
                merger_stretch,
                t_initial,
                t_merger,
                t_ringdown,
                t_final,
            )

        def transform_non_affine(self, a):
            return np.piecewise(
                a,
                [(a < self.t_merger), (a >= self.t_merger) & (a < self.t_ringdown), a >= self.t_ringdown],
                [
                    lambda x: x,
                    lambda x: self.t_merger + self.merger_stretch * (x - self.t_merger),
                    lambda x: (self.merger_stretch - 1) * (self.t_ringdown - self.t_merger) + x,
                ],
            )

        def inverted(self):
            return WaveformTimeScale.InvertedWaveformTimeTransform(
                self.merger_stretch, self.t_initial, self.t_merger, self.t_ringdown, self.t_final
            )

    class InvertedWaveformTimeTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True
        has_inverse = True

        def __init__(self, merger_stretch, t_initial, t_merger, t_ringdown, t_final):
            mtransforms.Transform.__init__(self)
            self.merger_stretch, self.t_initial, self.t_merger, self.t_ringdown, self.t_final = (
                merger_stretch,
                t_initial,
                t_merger,
                t_ringdown,
                t_final,
            )

        def transform_non_affine(self, f_a):
            return np.piecewise(
                f_a,
                [
                    (f_a < self.t_merger),
                    (
                        (f_a >= (1 - self.merger_stretch) * self.t_merger + self.merger_stretch * self.t_merger)
                        & (f_a < (1 - self.merger_stretch) * self.t_merger + self.merger_stretch * self.t_ringdown)
                    ),
                    f_a >= self.merger_stretch * self.t_ringdown - (self.merger_stretch - 1) * (self.t_merger),
                ],
                [
                    lambda x: x,
                    lambda x: (x - (1 - self.merger_stretch) * self.t_merger) / self.merger_stretch,
                    lambda x: x - (self.merger_stretch - 1) * (self.t_ringdown - self.t_merger),
                ],
            )

        def inverted(self):
            return WaveformTimeScale.WaveformTimeTransform(
                self.merger_stretch, self.t_initial, self.t_merger, self.t_ringdown, self.t_final
            )


mscale.register_scale(WaveformTimeScale)



import numpy as np
from spherical_functions import LM_total_size
from .. import WaveformModes
from .. import ModesTimeSeries
from .. import Inertial
from .. import h as h_DataType


class AsymptoticBondiData:
    """Class to store asymptotic Bondi data

    This class stores time data, along with the corresponding values of psi0 through psi4 and sigma.
    For simplicity, the data are stored as one contiguous array.  That is, *all* values are stored
    at all times, even if they are zero, and all Modes objects are stored with ell_min=0, even when
    their spins are not zero.

    The single contiguous array is then viewed as 6 separate ModesTimeSeries objects, which enables
    them to track their spin weights, and provides various convenient methods like `eth` and
    `ethbar`; `dot` and `ddot` for time-derivatives; `int` and `iint` for time-integrations; `norm`
    to take the norm of a function over the sphere; `bar` for conjugation of the functions (which is
    different from just conjugating the mode weights); etc.  It also handles algebra correctly --
    particularly addition (which is disallowed when the spin weights differ) and multiplication
    (which can be delicate with regards to the resulting ell values).

    This may lead to some headaches when the user tries to do things that are disabled by Modes
    objects.  The goal is to create headaches if and only if the user is trying to do things that
    really should never be done (like conjugating mode weights, rather than the underlying function;
    adding modes with different spin weights; etc.).  Please open issues for any situations that
    don't meet this standard.

    This class also provides various convenience methods for computing things like the mass aspect,
    the Bondi four-momentum, the Bianchi identities, etc.

    """

    def __init__(self, time, ell_max, multiplication_truncator=sum, frameType=Inertial):
        """Create new storage for asymptotic Bondi data

        Parameters
        ==========
        time: int or array_like
            Times at which the data will be stored.  If this is an int, an empty array of that size
            will be created.  Otherwise, this must be a 1-dimensional array of floats.
        ell_max: int
            Maximum ell value to be stored
        multiplication_truncator: callable [defaults to `sum`, even though `max` is nicer]
            Function to be used by default when multiplying Modes objects together.  See the
            documentation for spherical_functions.Modes.multiply for more details.  The default
            behavior with `sum` is the most correct one -- keeping all ell values that result -- but
            also the most wasteful, and very likely to be overkill.  The user should probably always
            use `max`.  (Unfortunately, this must remain an opt-in choice, to ensure that the user
            is aware of the situation.)

        """
        import functools

        if np.ndim(time) == 0:
            # Assume this is just the size of the time array; construct an empty array
            time = np.empty((time,), dtype=float)
        elif np.ndim(time) > 1:
            raise ValueError(f"Input `time` parameter must be an integer or a 1-d array; it has shape {time.shape}")
        if time.dtype != float:
            raise ValueError(f"Input `time` parameter must have dtype float; it has dtype {time.dtype}")
        ModesTS = functools.partial(ModesTimeSeries, ell_max=ell_max, multiplication_truncator=multiplication_truncator)
        shape = [6, time.size, LM_total_size(0, ell_max)]
        self.frame = np.array([])
        self.frameType = frameType
        self._time = time.copy()
        self._raw_data = np.zeros(shape, dtype=complex)
        self._psi0 = ModesTS(self._raw_data[0], self._time, spin_weight=2)
        self._psi1 = ModesTS(self._raw_data[1], self._time, spin_weight=1)
        self._psi2 = ModesTS(self._raw_data[2], self._time, spin_weight=0)
        self._psi3 = ModesTS(self._raw_data[3], self._time, spin_weight=-1)
        self._psi4 = ModesTS(self._raw_data[4], self._time, spin_weight=-2)
        self._sigma = ModesTS(self._raw_data[5], self._time, spin_weight=2)

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, new_time):
        self._time[:] = new_time
        return self._time

    u = time

    t = time

    @property
    def n_times(self):
        return self.time.size

    @property
    def n_modes(self):
        return self._raw_data.shape[-1]

    @property
    def ell_min(self):
        return self._psi2.ell_min

    @property
    def ell_max(self):
        return self._psi2.ell_max

    @property
    def LM(self):
        return self.psi2.LM

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, sigmaprm):
        self._sigma[:] = sigmaprm
        return self.sigma

    @property
    def h(self):
        h_mts = 2.0 * self._sigma.bar
        return WaveformModes(
            t=h_mts.t,
            data=np.array(h_mts)[:, h_mts.index(abs(h_mts.s), -abs(h_mts.s)) :],
            ell_min=abs(h_mts.s),
            ell_max=h_mts.ell_max,
            frameType=Inertial,
            dataType=h_DataType,
            r_is_scaled_out=True,
            m_is_scaled_out=True,
        )

    @property
    def psi4(self):
        return self._psi4

    @psi4.setter
    def psi4(self, psi4prm):
        self._psi4[:] = psi4prm
        return self.psi4

    @property
    def psi3(self):
        return self._psi3

    @psi3.setter
    def psi3(self, psi3prm):
        self._psi3[:] = psi3prm
        return self.psi3

    @property
    def psi2(self):
        return self._psi2

    @psi2.setter
    def psi2(self, psi2prm):
        self._psi2[:] = psi2prm
        return self.psi2

    @property
    def psi1(self):
        return self._psi1

    @psi1.setter
    def psi1(self, psi1prm):
        self._psi1[:] = psi1prm
        return self.psi1

    @property
    def psi0(self):
        return self._psi0

    @psi0.setter
    def psi0(self, psi0prm):
        self._psi0[:] = psi0prm
        return self.psi0

    def copy(self):
        import copy

        new_abd = type(self)(self.t, self.ell_max)
        state = copy.deepcopy(self.__dict__)
        new_abd.__dict__.update(state)
        return new_abd

    def interpolate(self, new_times):
        new_abd = type(self)(new_times, self.ell_max)
        new_abd.frameType = self.frameType
        # interpolate waveform data
        new_abd.sigma = self.sigma.interpolate(new_times)
        new_abd.psi4 = self.psi4.interpolate(new_times)
        new_abd.psi3 = self.psi3.interpolate(new_times)
        new_abd.psi2 = self.psi2.interpolate(new_times)
        new_abd.psi1 = self.psi1.interpolate(new_times)
        new_abd.psi0 = self.psi0.interpolate(new_times)
        # interpolate frame data if necessary
        if self.frame.shape[0] == self.n_times:
            import quaternion

            new_abd.frame = quaternion.squad(self.frame, self.t, new_times)
        return new_abd

    from .from_initial_values import from_initial_values

    from .constraints import (
        bondi_constraints,
        bondi_violations,
        bondi_violation_norms,
        bianchi_0,
        bianchi_1,
        bianchi_2,
        constraint_3,
        constraint_4,
        constraint_mass_aspect,
    )

    from .transformations import transform
    from .bms_charges import (
        mass_aspect,
        bondi_rest_mass,
        bondi_four_momentum,
        bondi_angular_momentum,
        CWWY_angular_momentum,
        bondi_dimensionless_spin,
        bondi_boost_charge,
        bondi_CoM_charge,
        supermomentum,
    )

    from .map_to_superrest_frame import map_to_superrest_frame
    from .map_to_abd_frame import map_to_abd_frame
